# -*- coding: utf-8 -*-
#  Copyright (c) 2020. Distributed under the terms of the MIT License.
import numpy as np
import pytest

from pydefect.defaults import defaults
from pydefect.input_maker.defect import SimpleDefect
from pydefect.input_maker.defect_entries_maker import (
    DefectEntriesMaker, copy_to_structure, to_istructure, add_atom_to_structure,
    perturb_structure, random_3d_vector)
from pydefect.input_maker.defect_entry import DefectEntry, PerturbedSite
from pydefect.input_maker.defect_set import DefectSet
from pydefect.input_maker.supercell_info import SupercellInfo, Site, \
    Interstitial
from vise.tests.helpers.assertion import assert_dataclass_almost_equal


@pytest.fixture
def cubic_supercell_info(cubic_supercell):
    sites = {"H1": Site(element="H", wyckoff_letter="a", site_symmetry="m-3m",
                        equivalent_atoms=list(range(32))),
             "He1": Site(element="He", wyckoff_letter="b", site_symmetry="m-3m",
                         equivalent_atoms=list(range(32, 64)))}
    interstitial = Interstitial([0.25]*3, site_symmetry="yy")
    return SupercellInfo(cubic_supercell,
                         "Fm-3m",
                         [[2, 0, 0], [0, 2, 0], [0, 0, 2]],
                         sites,
                         interstitials=[interstitial])


def test_defect_entries_maker(cubic_supercell_info, cubic_supercell, mocker):
    mock = mocker.patch("pydefect.input_maker.defect_entries_maker.defaults")
    mock.displace_distance = 0.0
    defect_set = DefectSet(defects={SimpleDefect(None, "H1", [1]),
                                    SimpleDefect("Ne", "He1", [0])})
    maker = DefectEntriesMaker(cubic_supercell_info, defect_set)

    va_h1_str = cubic_supercell.copy()
    va_h1_str.pop(0)
    va_h1_str = to_istructure(va_h1_str)

    ne_he1_str = cubic_supercell.copy()
    coords = ne_he1_str.pop(32).frac_coords
    ne_he1_str.append("Ne", coords)
    ne_he1_str = to_istructure(ne_he1_str)

    defect_entries = {DefectEntry(name="Ne_He1",
                                  charge=0,
                                  structure=ne_he1_str,
                                  site_symmetry="m-3m",
                                  defect_center=(0.25, 0.0, 0.0)),
                      DefectEntry(name="Va_H1",
                                  charge=1,
                                  structure=va_h1_str,
                                  site_symmetry="m-3m",
                                  defect_center=(0.0, 0.0, 0.0))}

    assert maker.defect_entries == defect_entries


def test_defect_entries_maker_insert_host_atoms(cubic_supercell_info, mocker):
    mock = mocker.patch("pydefect.input_maker.defect_entries_maker.defaults")
    mock.displace_distance = 0.0

    defect_set = DefectSet(defects={SimpleDefect("He", "H1", [0]),
                                    SimpleDefect("H", "He1", [0])})
    maker = DefectEntriesMaker(cubic_supercell_info, defect_set)

    for defect_entry in maker.defect_entries:
        if defect_entry.name == "H_He1":
            actual = defect_entry.structure[0].frac_coords
            np.testing.assert_array_equal(actual, [0.25, 0.0, 0.0])
        if defect_entry.name == "He_H1":
            actual = defect_entry.structure[31].frac_coords
            np.testing.assert_array_equal(actual, [0.0, 0.0, 0.0])


def test_defect_entries_interstitials(cubic_supercell_info, mocker):
    mock = mocker.patch("pydefect.input_maker.defect_entries_maker.defaults")
    mock.displace_distance = 0.0

    defect_set = DefectSet(defects={SimpleDefect("H", "i1", [0])})
    maker = DefectEntriesMaker(cubic_supercell_info, defect_set)

    defect_entry = maker.defect_entries.pop()
    actual = defect_entry.structure[0].frac_coords
    np.testing.assert_array_equal(actual, [0.25, 0.25, 0.25])


def test_add_atom_to_structure(ortho_conventional):
    coords = [0.1, 0.1, 0.1]

    structure_1 = copy_to_structure(ortho_conventional)
    add_atom_to_structure(structure_1, "H", coords)
    assert len(structure_1) == 9
    np.testing.assert_array_equal(structure_1[0].frac_coords, coords)

    structure_2 = copy_to_structure(ortho_conventional)
    add_atom_to_structure(structure_2, "He", coords)
    assert len(structure_2) == 9
    np.testing.assert_array_equal(structure_2[4].frac_coords, coords)

    structure_3 = copy_to_structure(ortho_conventional)
    add_atom_to_structure(structure_3, "Li", coords)
    assert len(structure_3) == 9
    np.testing.assert_array_equal(structure_3[8].frac_coords, coords)


def test_perturbed_structure(ortho_conventional, mocker):
    mock = mocker.patch(
        "pydefect.input_maker.defect_entries_maker.random_3d_vector")
    mock.return_value = ([0.0, 0.0, 0.35], 1.0)
    structure = copy_to_structure(ortho_conventional)
    actual = perturb_structure(structure,
                               center=[0.0, 0.0, 0.0], cutoff=2.5-1e-5)
    expected_structure = structure
    expected_structure[0].frac_coords = [0.0, 0.0, 0.05]
    expected_site = PerturbedSite(element='H',
                                  distance=0.0,
                                  initial_coords=(0.0, 0.0, 0.0),
                                  perturbed_coords=(0.0, 0.0, 0.05),
                                  displacement=1.0)

    mock.assert_called_with(defaults.displace_distance)
    assert actual[0] == expected_structure
    assert_dataclass_almost_equal(actual[1][0], expected_site)


def test_random_3d_vector():
    actual_vec, actual_dist = random_3d_vector(1.0)
    assert actual_dist <= 1.0
    assert len(actual_vec) == 3

    actual_vec, actual_dist = random_3d_vector(0.0)
    assert actual_dist == 0.0

"""
TODO
- Set cutoff  properly.

DONE
- Take SupercellInfo and DefectSet, return List[DefectEntry]
- Insert host antisite atoms to the middle of structure 
"""


# -*- coding: utf-8 -*-
#  Copyright (c) 2020 Kumagai group.

from pymatgen.io.vasp.inputs import Poscar, Incar, Potcar
from vise.util.logger import get_logger

logger = get_logger(__name__)


def get_defect_charge_state(poscar: Poscar, potcar: Potcar, incar: Incar):
    """Get defect charge state from structure, potcar, and NELECT in INCAR. """
    nelect = incar.get("NELECT", None)
    if nelect is None:
        logger.info("Since NELECT is not written in INCAR, so 0 is returned.")
        return 0
    potcar_elements = [symbol.split("_")[0] for symbol in potcar.symbols]
    if poscar.site_symbols != potcar_elements:
        raise ValueError(f"Sequence of elements in POSCAR {poscar.site_symbols}"
                         f" and that in POTCAR {potcar.symbols} is different.")
    num_elect_neutral = sum([num_atom * pot.nelectrons
                             for num_atom, pot in zip(poscar.natoms, potcar)])
    excess_num_electrons = int(nelect - num_elect_neutral)
    return - excess_num_electrons





# -*- coding: utf-8 -*-
#  Copyright (c) 2020 Kumagai group.

from pydefect.analyzer.defect_energy import DefectEnergyInfo, DefectEnergy, \
    DefectEnergySummary, DefectEnergies
from pydefect.analyzer.make_defect_energy_summary import \
    make_defect_energy_summary
from pydefect.chem_pot_diag.chem_pot_diag import TargetVertices, TargetVertex


def test_make_defect_energy_summary(mocker):
    energy1 = DefectEnergy(0.0, {"PC correction": 2.0}, False)
    energy2 = DefectEnergy(1.0, {"PC correction": 3.0}, True)
    defect_infos = [DefectEnergyInfo("Va_Mg1", 0, {"Mg": -1}, energy1),
                    DefectEnergyInfo("Va_Mg1", 1, {"Mg": -1}, energy2)]
    target_vertices = TargetVertices(
        target="MgO", vertices={"A": TargetVertex({"Mg": 5.0})})

    unitcell = mocker.Mock()
    unitcell.vbm = 1.0
    unitcell.cbm = 11.0

    perf_be_state = mocker.Mock()
    perf_be_state.vbm_info.energy = 0.0
    perf_be_state.cbm_info.energy = 12.0

    actual = make_defect_energy_summary(defect_infos, target_vertices, unitcell,
                                        perf_be_state)

    defect_energies = {"Va_Mg1": DefectEnergies(
        atom_io={"Mg": -1}, charges=[0, 1], defect_energies=[energy1, energy2])}

    expected = DefectEnergySummary(title=unitcell.system,
                                   defect_energies=defect_energies,
                                   rel_chem_pots={"A": {"Mg": 5.0}},
                                   cbm=10.0,
                                   supercell_vbm=-1.0,
                                   supercell_cbm=11.0)
    assert actual == expected



# -*- coding: utf-8 -*-
#  Copyright (c) 2020 Kumagai group.
import math
import warnings
from typing import List, Tuple

import numpy as np
from pydefect.analyzer.defect_structure_comparator import \
    DefectStructureComparator
from pydefect.analyzer.defect_structure_info import Displacement, \
    DefectStructureInfo, unique_point_group
from pydefect.defaults import defaults
from pymatgen.core import PeriodicSite, Structure
from vise.util.structure_symmetrizer import StructureSymmetrizer
from vise.util.typing import GenCoords


def folded_coords(site: PeriodicSite,
                  center: GenCoords
                  ) -> GenCoords:
    _, image = site.distance_and_image_from_frac_coords(center)
    return tuple(site.frac_coords - image)


class MakeDefectStructureInfo:
    def __init__(self,
                 perfect: Structure,
                 initial: Structure,
                 final: Structure,
                 symprec: float,
                 dist_tol: float,
                 neighbor_cutoff_factor: float = None):

        self.cutoff = neighbor_cutoff_factor or defaults.cutoff_distance_factor
        self.symprec = symprec
        self.perfect, self.initial, self.final = perfect, initial, final
        self.dist_tol = dist_tol

        assert perfect.lattice == initial.lattice == final.lattice
        self.lattice = perfect.lattice

        self._orig_comp = DefectStructureComparator(final, perfect, dist_tol)
        self._orig_center = self._orig_comp.defect_center_coord
        self._calc_drift()

        self.shifted_final = final.copy()
        self.center = tuple(self._orig_center - self._drift_vector)
        for site in self.shifted_final:
            site.frac_coords -= np.array(self._drift_vector)

        self.comp_w_perf = DefectStructureComparator(
            self.shifted_final, perfect, dist_tol)
        self.comp_w_init = DefectStructureComparator(
            self.shifted_final, initial, dist_tol)

        self.defect_structure_info = DefectStructureInfo(
            shifted_final_structure=self.shifted_final,
            initial_site_sym=self.initial_site_sym,
            final_site_sym=self.final_site_sym,
            site_diff=self.comp_w_perf.make_site_diff(),
            site_diff_from_initial=self.comp_w_init.make_site_diff(),
            symprec=symprec,
            dist_tol=dist_tol,
            anchor_atom_idx=self._anchor_atom_idx,
            neighbor_atom_indices=self._neighbor_atom_indices,
            neighbor_cutoff_factor=self.cutoff,
            drift_vector=self._drift_vector,
            drift_dist=self._drift_distance,
            center=self.center,
            displacements=self.calc_displacements())

    @property
    def _neighbor_atom_indices(self):
        return self.comp_w_perf.neighboring_atom_indices(self.cutoff)

    @property
    def initial_site_sym(self):
        return self._unique_point_group(self.initial)

    @property
    def final_site_sym(self):
        return self._unique_point_group(self.final)

    def _unique_point_group(self, structure):
        symmetrizer = StructureSymmetrizer(structure, self.symprec)
        return unique_point_group(symmetrizer.point_group)

    def _calc_drift(self) -> None:
        distances = []
        for site in self.final:
            distances.append(
                site.distance_and_image_from_frac_coords(self._orig_center)[0])

        self._anchor_atom_idx = int(np.argmax(distances))
        p_anchor_atom_idx = self._orig_comp.d_to_p[self._anchor_atom_idx]
        d_site = self.final[self._anchor_atom_idx]
        p_coords = self.perfect[p_anchor_atom_idx].frac_coords
        self._drift_distance, image = \
            d_site.distance_and_image_from_frac_coords(p_coords)
        self._drift_vector = tuple(d_site.frac_coords - p_coords - image)

    def calc_displacements(self):
        result = []
        atom_mapping = self.comp_w_init.atom_mapping
        for d in range(len(self.shifted_final)):
            if d not in atom_mapping:
                result.append(None)
            else:
                p = atom_mapping[d]
                p_elem = str(self.initial[p].specie)
                d_elem = str(self.shifted_final[d].specie)
                if p_elem != d_elem:
                    result.append(None)
                    continue
                initial_pos = folded_coords(self.initial[p], self.center)
                final_pos = folded_coords(self.shifted_final[d], initial_pos)

                initial_pos_vec = self.lattice.get_cartesian_coords(
                    np.array(initial_pos) - self.center)
                initial_dist = np.linalg.norm(initial_pos_vec)

                disp_dist, t = self.lattice.get_distance_and_image(
                    self.shifted_final.frac_coords[d], self.initial.frac_coords[p])
                disp_vec = self.lattice.get_cartesian_coords(
                    self.shifted_final.frac_coords[d] - self.initial.frac_coords[p] - t)

                angle = self.calc_disp_angle(disp_dist, disp_vec, initial_dist,
                                             initial_pos_vec)

                result.append(Displacement(specie=p_elem,
                                           original_pos=initial_pos,
                                           final_pos=final_pos,
                                           distance_from_defect=initial_dist,
                                           disp_vector=tuple(disp_vec),
                                           displace_distance=disp_dist,
                                           angle=angle))
        return result

    @staticmethod
    def calc_disp_angle(disp_dist, disp_vec, ini_dist, initial_pos_vec):
        inner_prod = sum(initial_pos_vec * disp_vec)
        with warnings.catch_warnings():
            warnings.simplefilter("ignore")
            # ignore "RuntimeWarning: invalid value encountered in double_scalars"
            cos = round(inner_prod / (ini_dist * disp_dist), 10)
        result = float(round(180 * (1 - np.arccos(cos) / np.pi), 1))
        if math.isnan(result):
            result = None
        return result




# -*- coding: utf-8 -*-
#  Copyright (c) 2020. Distributed under the terms of the MIT License.

from pydefect.input_maker.defect import (
    Defect, SimpleDefect, screen_simple_defect)
from vise.tests.helpers.assertion import assert_msonable


def test_defect_msonable():
    defect = Defect("Va_O1", (1, 2))
    assert_msonable(defect)


def test_defect_hashable():
    defect = Defect("Va_O1", (1, 2))
    d = {defect: 1}


def test_simple_defect_hashable():
    defect = SimpleDefect(None, "O1", (1, 2))
    d = {defect: 1}


def test_defect_name_and_charges():
    defect = Defect("Va_O1", (1, 2))
    assert defect.str_list == ["Va_O1_1", "Va_O1_2"]


def test_simple_defect():
    defect = SimpleDefect(None, "O1", [1, 2])
    expected = Defect("Va_O1", (1, 2))
    assert defect.name == expected.name
    assert defect.charge_list == expected.charge_list
    assert defect.in_atom is None


def test_screen_defect():
    defect = SimpleDefect(None, "O1", [1, 2])
    assert screen_simple_defect(defect, ["Va_O1"]) == \
           SimpleDefect(None, "O1", [1, 2])
    assert screen_simple_defect(defect, ["Va_O1_1"]) == \
           SimpleDefect(None, "O1", [1])
    assert screen_simple_defect(defect, ["Va_N1"]) is None




# -*- coding: utf-8 -*-
#  Copyright (c) 2020. Distributed under the terms of the MIT License.
from argparse import Namespace
from pathlib import Path

from pydefect.analyzer.calc_results import CalcResults
from pydefect.chem_pot_diag.chem_pot_diag import CompositionEnergies, \
    CompositionEnergy
from pydefect.cli.main_util_functions import composition_energies_from_mp, \
    make_gkfo_correction_from_vasp, add_interstitials_from_local_extrema, \
    make_defect_vesta_file
from pydefect.corrections.efnv_correction import ExtendedFnvCorrection
from pymatgen.core import Composition


def test_composition_energies_from_mp(mocker, tmpdir):
    tmpdir.chdir()
    args = Namespace(elements=["H"], atom_energy_yaml="a.yaml")
    mock = mocker.patch(
        "pydefect.cli.main_util_functions.make_composition_energies_from_mp")
    mock.return_value = CompositionEnergies(
        {Composition("H"): CompositionEnergy(1.0, "mp-1")})
    composition_energies_from_mp(args)
    mock.assert_called_once_with(elements=["H"], atom_energy_yaml="a.yaml")
    actual = Path("composition_energies.yaml").read_text()
    expected = """H1:
  energy: 1.0
  source: mp-1
"""
    assert actual == expected


def test_add_interstitials_from_local_extrema(mocker):
    mock_supercell_info = mocker.Mock()
    mock_local_extrema = mocker.Mock()
    args = Namespace(supercell_info=mock_supercell_info,
                     local_extrema=mock_local_extrema,
                     indices=[1, 2])
    add_interstitials_from_local_extrema(args)
    mock_local_extrema.append_sites_to_supercell_info.assert_called_once_with(
        mock_supercell_info, [1, 2])


def test_make_defect_vesta_file(mocker):
    mock_defect_structure_info = mocker.Mock()

    def side_effect(key):
        if str(key) == "Va_O1_0/defect_structure_info.json":
            return mock_defect_structure_info
        else:
            print(key)
            raise ValueError

    mock_loadfn = mocker.patch("pydefect.cli.main_util_functions.loadfn",
                               side_effect=side_effect)
    mock_loadfn2 = mocker.patch("pydefect.cli.main_functions.loadfn",
                                side_effect=side_effect)
    mock_make_vesta_file = mocker.patch(
        "pydefect.cli.main_util_functions.MakeDefectVestaFile")
    args = Namespace(
        dirs=[Path("Va_O1_0")],
        check_calc_results=True,
        cutoff=1.0,
        min_displace_w_arrows=2.0,
        arrow_factor=3.0,
        title=None)
    make_defect_vesta_file(args)
    mock_make_vesta_file.assert_called_once_with(mock_defect_structure_info,
                                                 1.0, 2.0, 3.0, None)


def test_make_gkfo_correction_from_vasp(tmpdir, mocker):
    mock_i_correction = mocker.Mock(spec=ExtendedFnvCorrection, autospec=True)
    mock_i_calc_results = mocker.Mock(spec=CalcResults, autospec=True)
    mock_f_calc_results = mocker.Mock(spec=CalcResults, autospec=True)
    mock_unitcell = mocker.Mock()

    mock_pot_plotter = mocker.patch(
        "pydefect.cli.main_util_functions.SitePotentialMplPlotter")
    mock_make_gkfo = mocker.patch(
        "pydefect.cli.main_util_functions.make_gkfo_correction")

    args = Namespace(
        initial_efnv_correction=mock_i_correction,
        initial_calc_results=mock_i_calc_results,
        final_calc_results=mock_f_calc_results,
        charge_diff=1,
        unitcell=mock_unitcell)

    make_gkfo_correction_from_vasp(args)
    mock_make_gkfo.assert_called_with(
        efnv_correction=mock_i_correction,
        additional_charge=1,
        final_calc_results=mock_f_calc_results,
        initial_calc_results=mock_i_calc_results,
        diele_tensor=mock_unitcell.dielectric_constant,
        ion_clamped_diele_tensor=mock_unitcell.ele_dielectric_const)





# -*- coding: utf-8 -*-
#  Copyright (c) 2020. Distributed under the terms of the MIT License.
import string
from copy import deepcopy
from dataclasses import dataclass, asdict
from itertools import product
from typing import Dict, Optional, Union, List, Set, Tuple

import numpy as np
import yaml
from monty.json import MSONable
from monty.serialization import loadfn
from pydefect.error import PydefectError
from pymatgen.core import Composition
from scipy.spatial.qhull import HalfspaceIntersection
from vise.util.logger import get_logger
from vise.util.mix_in import ToYamlFileMixIn, ToJsonFileMixIn

AtoZ = list(string.ascii_uppercase)
LargeMinusNumber = -1e5

logger = get_logger(__name__)


@dataclass
class CompositionEnergy(MSONable):
    energy: float
    source: str = None


class CompositionEnergies(ToYamlFileMixIn, dict):
    """
    keys: Composition
    values: CompositionEnergy
    """
    def to_yaml(self) -> str:
        d = {}
        for k, v in self.items():
            key = str(k.iupac_formula).replace(" ", "")
            val = {"energy": v.energy, "source": str(v.source)}
            d[key] = val
        return yaml.dump(d)

    @classmethod
    def from_yaml(cls, filename: str = None):
        name = filename or cls._yaml_filename()
        d = loadfn(name)
        composition_energies = {}
        for k, v in d.items():
            source = v.get('source', None)
            key = Composition(k)
            composition_energies[key] = CompositionEnergy(v["energy"], source)
        return cls(composition_energies)

    @classmethod
    def from_dict(cls, d: Dict[str, float]):
        ce = {Composition(k): CompositionEnergy(v) for k, v in d.items()}
        return cls(ce)

    @property
    def elements(self):
        result = set()
        for c in self:
            result.update(set([str(e) for e in c.elements]))
        return sorted(result)

    @property
    def std_rel_energies(self) -> Tuple["StandardEnergies", "RelativeEnergies"]:
        std, rel = StandardEnergies(), RelativeEnergies()
        abs_energies_per_atom = {k.reduced_formula: v.energy / k.num_atoms
                                 for k, v in self.items()}
        std_energies_list = []
        for vertex_element in self.elements:
            # This target is needed as some reduced formulas shows molecule
            # ones such as H2 and O2.
            reduced_formula = Composition({vertex_element: 1.0}).reduced_formula
            candidates = filter(lambda x: x[0] == reduced_formula,
                                abs_energies_per_atom.items())
            try:
                min_abs_energy = min([abs_energy_per_atom[1]
                                      for abs_energy_per_atom in candidates])
            except ValueError:
                print(f"Element {vertex_element} does not exist in "
                      f"CompositionEnergies.")
                raise NoElementEnergyError
            std[vertex_element] = min_abs_energy
            std_energies_list.append(min_abs_energy)

        for formula, abs_energy_per_atom in abs_energies_per_atom.items():
            if Composition(formula).is_element:
                continue
            frac = atomic_fractions(formula, self.elements)
            offset = sum([a * b for a, b in zip(frac, std_energies_list)])
            rel[formula] = abs_energy_per_atom - offset

        return std, rel


class CpdAbstractEnergies(ToYamlFileMixIn, dict):
    """
    keys: str (composition name)
    values: float (energy per atom)
    """
    def to_yaml(self) -> str:
        return yaml.dump(dict(self))

    @classmethod
    def from_yaml(cls, filename: str = None):
        return cls(loadfn(filename or cls._yaml_filename()))


class StandardEnergies(CpdAbstractEnergies):
    pass


def atomic_fractions(comp: Union[Composition, str], elements: List[str]
                     ) -> List[float]:
    return [Composition(comp).fractional_composition[e] for e in elements]


def comp_to_element_set(comp: Union[Composition, str]) -> Set[str]:
    return {str(e) for e in Composition(comp).elements}


def target_element_chem_pot(comp: Union[Composition, str],
                            energy_per_atom: float,
                            target_element: str,
                            other_elem_chem_pot: Dict[str, float]) -> float:
    assert comp_to_element_set(comp) \
           <= set(other_elem_chem_pot) | {target_element}
    other_element_val = 0.0
    for element, frac in Composition(comp).fractional_composition.items():
        if target_element == str(element):
            target_frac = frac
            continue
        other_element_val += other_elem_chem_pot[str(element)] * frac
    return (energy_per_atom - other_element_val) / target_frac


class RelativeEnergies(CpdAbstractEnergies):
    #TODO: add stability
    @property
    def all_element_set(self) -> Set[str]:
        return set().union(*[comp_to_element_set(c) for c in self])

    def host_composition_energies(self, elements: List[str]
                                  ) -> Dict[str, float]:
        return {formula: energy for formula, energy in self.items()
                if comp_to_element_set(formula).issubset(elements)}

    def comp_energies_with_element(self,
                                   element: str) -> Dict[str, float]:
        return {formula: energy for formula, energy in self.items()
                if element in comp_to_element_set(formula)}

    def impurity_chem_pot(self, impurity_element: str,
                          host_elements_chem_pot: Dict[str, float]
                          ) -> Tuple[float, str]:
        impurity_chem_pot = {impurity_element: 0.0}
        comp_energies = self.comp_energies_with_element(impurity_element)

        for formula, energy_per_atom in comp_energies.items():
            impurity_chem_pot[formula] = target_element_chem_pot(
                formula, energy_per_atom, impurity_element,
                host_elements_chem_pot)

        competing_phase_formula = min(impurity_chem_pot,
                                      key=impurity_chem_pot.get)
        chem_pot = impurity_chem_pot[competing_phase_formula]
        return chem_pot, competing_phase_formula


class ChemPotDiagMaker:
    def __init__(self,
                 relative_energies: RelativeEnergies,
                 elements: List[str],
                 target: str = None):
        self.relative_energies = relative_energies
        self.host_composition_energies = \
            relative_energies.host_composition_energies(elements)
        self.elements = elements
        self.impurity_elements = \
            relative_energies.all_element_set.difference(elements)
        self.dim = len(elements)

        if target:
            try:
                assert target in relative_energies.keys()
            except AssertionError:
                print(f"Target {target} is not in relative energy compounds.")
                raise
        self.target = target

    def _calc_vertices(self):
        half_spaces = []

        for formula, energy in self.host_composition_energies.items():
            half_spaces.append(
                atomic_fractions(formula, self.elements) + [-energy])

        for i in range(self.dim):
            upper_boundary, lower_boundary = [0.0] * self.dim, [0.0] * self.dim
            upper_boundary[i], lower_boundary[i] = 1.0, -1.0

            upper_boundary.append(0.0)
            lower_boundary.append(LargeMinusNumber)
            half_spaces.extend([upper_boundary, lower_boundary])

        feasible_point = np.array([LargeMinusNumber + 1.0] * self.dim,
                                  dtype=float)
        hs = HalfspaceIntersection(np.array(half_spaces), feasible_point)
        self.vertices: List[List[float]] = hs.intersections.tolist()

    def _is_composition_involved(self,
                                 coord: List[float],
                                 composition: str,
                                 energy: float) -> bool:
        atom_frac = atomic_fractions(composition, self.elements)
        diff = sum([x * y for x, y in zip(atom_frac, coord)]) - energy
        # If this is too small, the creation of cpd may fail.
        return abs(diff) < 1e-3

    def _min_energy_range(self, mul: float = 1.1):
        vertex_values = [x for x in sum(self.vertices, [])
                         if x != LargeMinusNumber]
        return min(vertex_values) * mul

    @property
    def chem_pot_diag(self):
        self._calc_vertices()
        elem_energies = {element: 0.0 for element in self.elements}
        host_energies = dict(**self.host_composition_energies, **elem_energies)

        polygons = {}
        for comp, energy in host_energies.items():
            vertices = []
            for coord in self.vertices:
                if self._is_composition_involved(coord, comp, energy):
                    vertex = [round(c, ndigits=5) if c != LargeMinusNumber
                              else self._min_energy_range() for c in coord]
                    vertices.append(vertex)
            if vertices:
                polygons[comp] = vertices

        vertices = None
        if self.target:
            target_vertices = []
            for coord in polygons[self.target]:
                competing_phases = []
                for comp, vertices in polygons.items():
                    if comp == self.target:
                        continue
                    if coord in vertices:
                        competing_phases.append(comp)

                impurity_phases = []
                host_chem_pots = dict(zip(self.elements, coord))
                impurity_chem_pots = {}
                for i_element in self.impurity_elements:
                    i_chem_pot, i_phase = \
                        self.relative_energies.impurity_chem_pot(
                            i_element, host_chem_pots)
                    impurity_chem_pots[i_element] = i_chem_pot
                    impurity_phases.append(i_phase)
                chem_pot = dict(**host_chem_pots, **impurity_chem_pots)
                target_vertices.append(TargetVertex(chem_pot,
                                                    competing_phases,
                                                    impurity_phases))

            AtoZZ = product([""] + AtoZ, AtoZ)
            vertices = {"".join(next(AtoZZ)): v for v in target_vertices}

        return ChemPotDiag(vertex_elements=self.elements,
                           polygons=polygons,
                           target=self.target,
                           target_vertices_dict=vertices)


@dataclass
class TargetVertex(MSONable):
    chem_pot: Dict[str, float]
    competing_phases: Optional[List[str]] = None
    impurity_phases: Optional[List[str]] = None


@dataclass
class TargetVertices(ToYamlFileMixIn):
    target: str
    vertices: Dict[str, TargetVertex]

    @property
    def chem_pots(self) -> Dict[str, Dict[str, float]]:
        return {k: v.chem_pot for k, v in self.vertices.items()}

    def to_yaml(self) -> str:
        header = f"target: {self.target}"
        d = {k: asdict(v) for k, v in self.vertices.items()}
        return "\n".join([header, yaml.dump(d)])

    @classmethod
    def from_yaml(cls, filename: str = None):
        name = filename or cls._yaml_filename()
        d = loadfn(name)
        target = d.pop("target")
        vertices = {k: TargetVertex(**v) for k, v in d.items()}
        return cls(target=target, vertices=vertices)


@dataclass
class ChemPotDiag(MSONable, ToJsonFileMixIn):
    vertex_elements: List[str]
    polygons: Dict[str, List[List[float]]]
    target: str = None
    target_vertices_dict: Dict[str, TargetVertex] = None

    @property
    def target_coords(self):
        if self.target_vertices_dict:
            return {k: [v.chem_pot[e] for e in self.vertex_elements]
                    for k, v in self.target_vertices_dict.items()}

    @property
    def min_range(self):
        return np.min(sum(self.polygons.values(), []))

    @property
    def dim(self):
        return len(self.vertex_elements)

    @property
    def comp_centers(self):
        _polygons = {}
        return {c: np.average(np.array(v), axis=0).tolist()
                for c, v in self.polygons.items()}

    def atomic_fractions(self, composition: str):
        return [Composition(composition).get_atomic_fraction(e)
                for e in self.vertex_elements]

    @property
    def chemical_system(self) -> str:
        return "-".join([el for el in self.vertex_elements])

    @property
    def to_target_vertices(self):
        if self.target and self.target_vertices_dict:
            return TargetVertices(self.target, self.target_vertices_dict)
        logger.warning("Need to set target and target_vertices.")
        raise ValueError


def change_element_sequence(cpd: ChemPotDiag,
                            element_sequence: List[str] = None) -> ChemPotDiag:
    result = deepcopy(cpd)

    if element_sequence:
        if (set(cpd.vertex_elements) != set(element_sequence)
                or len(cpd.vertex_elements) != len(element_sequence)):
            raise ValueError(f"Original elements {cpd.vertex_elements}. "
                             f"Input elements {element_sequence}")
    else:
        if cpd.target:
            element_sequence = [str(e) for e in Composition(cpd.target).elements]
        else:
            return result

    element_indices = [cpd.vertex_elements.index(e) for e in element_sequence]
    for k, v in cpd.polygons.items():
        result.polygons[k] = [[vv[x] for x in element_indices] for vv in v]

    result.vertex_elements = element_sequence
    return result


class NoElementEnergyError(PydefectError):
    pass



# -*- coding: utf-8 -*-
#  Copyright (c) 2020. Distributed under the terms of the MIT License.
from argparse import Namespace
from pathlib import Path

import pytest
from pydefect.analyzer.band_edge_states import BandEdgeStates
from pydefect.analyzer.calc_results import CalcResults
from pydefect.cli.vasp.main_vasp_util_functions import make_parchg_dir, \
    make_refine_defect_poscar, calc_charge_state, \
    calc_grids, make_defect_charge_info_main, make_defect_entry_main
from pymatgen.core import Structure
from vise.input_set.incar import ViseIncar
import numpy as np

_filepath = "pydefect.cli.vasp.main_vasp_util_functions"


def test_calc_charge_state(mocker):
    mock_poscar = mocker.patch(f"{_filepath}.Poscar")
    mock_potcar = mocker.patch(f"{_filepath}.Potcar")
    mock_incar = mocker.patch(f"{_filepath}.Incar")
    mock_get_charge_state = mocker.patch(f"{_filepath}.get_defect_charge_state")
    mock_get_charge_state.return_value = 0

    args = Namespace(dir=Path("Va_O1_0"))
    calc_charge_state(args)

    mock_poscar.from_file.assert_called_once_with(Path("Va_O1_0/POSCAR"))
    mock_potcar.from_file.assert_called_once_with(Path("Va_O1_0/POTCAR"))
    mock_incar.from_file.assert_called_once_with(Path("Va_O1_0/INCAR"))
    mock_get_charge_state.assert_called_once_with(
        mock_poscar.from_file.return_value,
        mock_potcar.from_file.return_value,
        mock_incar.from_file.return_value)


def test_make_defect_entry_main(mocker):
    mock_charge_state = mocker.patch(f"{_filepath}.calc_charge_state")
    mock_charge_state.return_value = 0
    mock_structure = mocker.patch(f"{_filepath}.Structure")
    mock_make_defect_entry = mocker.patch(f"{_filepath}.make_defect_entry")
    mock_perfect = mocker.Mock()

    args = Namespace(dir=Path("Va_O1_0"), name="Va_O1", perfect=mock_perfect)
    make_defect_entry_main(args)
    mock_charge_state.assert_called_once_with(args)
    mock_structure.from_file.assert_called_once_with(Path("Va_O1_0/POSCAR"))
    mock_make_defect_entry.assert_called_once_with(
        name="Va_O1",
        charge=0,
        perfect_structure=mock_perfect,
        defect_structure=mock_structure.from_file.return_value)


def test_make_parchg_dir(tmpdir, mocker):
    print(tmpdir)
    tmpdir.chdir()
    Path("WAVECAR").touch()
    Path("INCAR").write_text("ALGO  =  Normal")
    Path("POSCAR").write_text("a")
    Path("POTCAR").write_text("b")
    Path("KPOINTS").write_text("c")

    band_edge_states = mocker.Mock(spec=BandEdgeStates, autospec=True)
    band_edge_states.band_indices_from_vbm_to_cbm = [1, 2]

    calc_results = mocker.Mock(spec=CalcResults, autospec=True)

    def side_effect(key):
        if str(key) == "band_edge_states.json":
            return band_edge_states
        elif str(key) == "calc_results.json":
            return calc_results
        else:
            raise ValueError

    mock_loadfn = mocker.patch(f"{_filepath}.loadfn", side_effect=side_effect)
    args = Namespace(dir=Path(tmpdir))

    with pytest.raises(FileNotFoundError):
        make_parchg_dir(args)

    Path("WAVECAR").write_text("d")
    make_parchg_dir(args)

    assert Path("parchg/POSCAR").read_text() == "a"
    assert Path("parchg/POTCAR").read_text() == "b"
    assert Path("parchg/KPOINTS").read_text() == "c"
    assert Path("parchg/WAVECAR").read_text() == "d"

    actual = ViseIncar.from_file("parchg/INCAR")
    expected = {"ALGO": "Normal", "LPARD": True, "LSEPB": True, "KPAR": 1,
                "IBAND": "2 3"}
    assert actual == expected


def test_refine_poscar(tmpdir, mocker, before_refine, after_refine):
    print(tmpdir)
    tmpdir.chdir()
    mock_defect_entry = mocker.Mock()
    mock_defect_entry.anchor_atom_index = 1
    mock_defect_entry.anchor_atom_coords = np.array([0.0, 0.5, 0.5])

    args = Namespace(structure=before_refine,
                     defect_entry=mock_defect_entry,
                     poscar_name="refined_POSCAR")
    make_refine_defect_poscar(args)

    assert Structure.from_file("refined_POSCAR") == after_refine


def test_calc_grids(mocker):
    mock_grids = mocker.patch(f"{_filepath}.Grids")
    args = Namespace(chgcar="CHGCAR")
    calc_grids(args)
    mock_grids.from_chgcar.assert_called_once_with("CHGCAR")
    mock_grids.from_chgcar.return_value.dump.assert_called_once_with()


def test_make_defect_charge_info_main(mocker):
    mock_chgcar = mocker.patch(f"{_filepath}.Chgcar")
    m_chgcar = mock_chgcar.from_file.return_value
    mock_make_charge_info = mocker.patch(f"{_filepath}.make_defect_charge_info")
    mock_charge_info = mock_make_charge_info.return_value
    mock_grids = mocker.Mock()
    args = Namespace(parchgs=["PARCHG.0189.ALLK"],
                     grids=mock_grids,
                     bin_interval=0.1)
    make_defect_charge_info_main(args)

    mock_make_charge_info.assert_called_once_with([m_chgcar], [188], 0.1, mock_grids)
    mock_chgcar.from_file.assert_called_once_with("PARCHG.0189.ALLK")
    mock_charge_info.to_json_file.assert_called_once_with()



# -*- coding: utf-8 -*-
#  Copyright (c) 2020. Distributed under the terms of the MIT License.
from pathlib import Path
from shutil import copyfile
from typing import List

import yaml
from pydefect.cli.vasp.molecules.molecules import MOLECULE_DATA
from pymatgen.core import Composition
from vise.util.logger import get_logger

mol_dir = Path(__file__).parent / "molecules"

logger = get_logger(__name__)


def make_poscars_from_query(materials_query: List[dict], path: Path) -> None:
    for query in materials_query:
        reduced_formula = Composition(query["full_formula"]).reduced_formula
        if reduced_formula in MOLECULE_DATA.keys():
            _make_molecular_directory(path, reduced_formula)
        else:
            _make_solid_directory(path, reduced_formula, query)


def _make_solid_directory(path: Path, reduced_formula: str, query: dict):
    dirname = path / f"{reduced_formula}_{query['task_id']}"
    logger.info(f"{dirname.name} is being created with a MP structure.")
    dirname.mkdir()
    query["structure"].to(filename=dirname / "POSCAR")
    d = {"total_magnetization": query["total_magnetization"],
         "band_gap": query["band_gap"],
         "data_source": query["task_id"]}
    prior_info = dirname / "prior_info.yaml"
    prior_info.write_text(yaml.dump(d), None)


def _make_molecular_directory(path: Path, reduced_formula: str):
    dirname = path / f"mol_{reduced_formula}"
    logger.info(
        f"{dirname.name} is being created with molecular data in pydefect.")
    if not dirname.exists():
        dirname.mkdir()
        copyfile(mol_dir / reduced_formula / "POSCAR",
                 dirname / "POSCAR")
        copyfile(mol_dir / reduced_formula / "prior_info.yaml",
                 dirname / "prior_info.yaml")



# -*- coding: utf-8 -*-
#  Copyright (c) 2020. Distributed under the terms of the MIT License.


import argparse
import sys
import warnings
from pathlib import Path

from pydefect.cli.main import epilog, description, add_sub_parser, dirs_parsers
from pydefect.cli.vasp.main_vasp_functions import make_defect_entries, \
    make_unitcell, make_competing_phase_dirs, \
    make_calc_results, \
    make_band_edge_orb_infos_and_eigval_plot, make_perfect_band_edge_state, \
    make_local_extrema, make_composition_energies
from pydefect.defaults import defaults
from pymatgen.io.vasp import Vasprun, Outcar, Chgcar
from pymatgen.io.vasp.inputs import UnknownPotcarWarning

warnings.simplefilter('ignore', UnknownPotcarWarning)


def parse_args_main_vasp(args):

    parser = argparse.ArgumentParser(epilog=epilog,
                                     description=description + """       
    The command is used for creating and analyzing VASP input and output 
    files.""")

    subparsers = parser.add_subparsers()
    perfect_band_edge_parser = add_sub_parser(argparse,
                                              name="perfect_band_edge_state")

    # -- unitcell ------------------------------------------------
    parser_unitcell = subparsers.add_parser(
        name="unitcell",
        description="Parse unitcell info from vasp output files and create "
                    "unitcell.yaml.",
        formatter_class=argparse.ArgumentDefaultsHelpFormatter,
        aliases=['u'])

    parser_unitcell.add_argument(
        "-vb", "--vasprun_band", required=True, type=Vasprun,
        help="vasprun.xml file of band structure calculation.")
    parser_unitcell.add_argument(
        "-ob", "--outcar_band", required=True, type=Outcar,
        help="OUTCAR file of band structure calculation.")
    parser_unitcell.add_argument(
        "-odc", "--outcar_dielectric_clamped", required=True, type=Outcar,
        help="OUTCAR file of ion-clamped dielectric constant calculation.")
    parser_unitcell.add_argument(
        "-odi", "--outcar_dielectric_ionic", required=True, type=Outcar,
        help="OUTCAR file for calculating dielectric constant of ionic "
             "contribution.")
    parser_unitcell.add_argument(
        "-n", "--name", type=str,
        help="System name used for plotting defect formation energies.")
    parser_unitcell.set_defaults(func=make_unitcell)

    # -- make_poscars ------------------------------------------------
    parser_make_poscars = subparsers.add_parser(
        name="make_poscars",
        description="Make directories and poscar files in these that are "
                    "retrieved via the Materials Project API.",
        formatter_class=argparse.ArgumentDefaultsHelpFormatter,
        aliases=['mp'])

    parser_make_poscars.add_argument(
        "-e", "--elements", required=True, nargs="+", type=str,
        help="Element names involved. ")
    parser_make_poscars.add_argument(
        "--e_above_hull", default=defaults.e_above_hull, type=float,
        help="Allowed energy above hull in eV/atom.")

    parser_make_poscars.set_defaults(func=make_competing_phase_dirs)

    # -- make_composition_energies ---------------------------------------------
    parser_make_composition_energies = subparsers.add_parser(
        name="make_composition_energies",
        description="Make composition energies from the directories in which "
                    "competing phases are calculated. ",
        formatter_class=argparse.ArgumentDefaultsHelpFormatter,
        parents=dirs_parsers,
        aliases=['mce'])

    parser_make_composition_energies.add_argument(
        "-y", "--yaml_file", type=str,
        help="composition_energies.yaml to be overwritten.")

    parser_make_composition_energies.set_defaults(
        func=make_composition_energies)

    # -- make_local_extrema ----------------------------------------------------
    parser_make_local_extrema = subparsers.add_parser(
        name="local_extrema",
        description="Make local volumetric_data_local_extrema.json file.",
        formatter_class=argparse.ArgumentDefaultsHelpFormatter,
        aliases=['le'])

    parser_make_local_extrema.add_argument(
        "-v", "--volumetric_data", type=Chgcar.from_file, required=True,
        nargs="+",
        help="File names such as CHGCAR or LOCPOT. When multiple files are "
             "provided, the summed data (e.g., AECCAR0 + AECCAR2) will be "
             "parsed.")
    parser_make_local_extrema.add_argument(
        "--find_max", dest="find_max", action="store_true",
        help="Set when local maxima are searched instead of local minima.")
    parser_make_local_extrema.add_argument(
        "-i", "--info", type=str,
        help="Information related to the parsed volumetric data written"
             "as the name in volumetric_data_local_extrema.json.")
    parser_make_local_extrema.add_argument(
        "--threshold_frac", default=None, type=float,
        help="""Optional fraction of extrema shown, which returns 
        `threshold_frac * tot_num_extrema` extrema fractional coordinates 
         based on highest/lowest intensity. 
         For example, there are 3 extrema with 0.1, 0.2, 0.3, and if 
         threshold_frac is set to 0.5, then those with 0.2 and 0.3 are shown 
         as the threshold is now 0.3 * 0.5 = 0.15.
         It takes from 0 to 1, and default is 1.""")
    parser_make_local_extrema.add_argument(
        "--threshold_abs", default=None, type=float,
        help="""Optional filter. When searching for local
        minima, intensity <= threshold_abs returns; when searching for
        local maxima, intensity >= threshold_abs returns.

        Caveat: threshold_abs and threshold_frac should be set exclusively.""")
    parser_make_local_extrema.add_argument(
        "--min_dist", type=float, metavar="Angstrom", default=0.5,
        help="""Used to remove the predicted sites that are too close to 
        *existing atoms* in the structure. The minimum distance that a vertex 
        needs to be from existing atoms. Set 0 when switch off this flag.""")
    parser_make_local_extrema.add_argument(
        "--tol", type=float, metavar="Angstrom", default=0.5,
        help="""Group interstitials that are too close together using a tol.
        Set 0 when switch off this flag.""")
    parser_make_local_extrema.add_argument(
        "--radius", type=float, default=0.4,
        help="Radius of sphere around each site to evaluate the average "
             "quantity.")

    parser_make_local_extrema.set_defaults(func=make_local_extrema)
    # -- defect_entries ------------------------------------------------
    parser_defect_entries = subparsers.add_parser(
        name="defect_entries",
        description="Make POSCAR files and defect entries in defect "
                    "directories",
        formatter_class=argparse.ArgumentDefaultsHelpFormatter,
        aliases=['de'])

    parser_defect_entries.set_defaults(func=make_defect_entries)

    # -- calc_results ------------------------------------------------
    parser_calc_results = subparsers.add_parser(
        name="calc_results",
        description="Generate calc_results.json files. "
                    "Run after VASP calculations.",
        parents=dirs_parsers,
        formatter_class=argparse.ArgumentDefaultsHelpFormatter,
        aliases=['cr'])
    parser_calc_results.set_defaults(func=make_calc_results)

    # -- perfect band edge state  ----------------------------------------------
    parser_perf_band_edge_state = subparsers.add_parser(
        name="perfect_band_edge_state",
        description="Generate band_edge_orbital_infos.json and "
                    "eigenvalues.pdf.",
        formatter_class=argparse.ArgumentDefaultsHelpFormatter,
        aliases=['pbes'])
    parser_perf_band_edge_state.add_argument(
        "-d", "--dir", type=Path,
        help="Directory path to the perfect supercell calculation.")

    parser_perf_band_edge_state.set_defaults(func=make_perfect_band_edge_state)

    # -- band edge orbital infos  ----------------------------------------------
    parser_band_edge_orb_infos = subparsers.add_parser(
        name="band_edge_orbital_infos",
        description="Generate band_edge_orbital_infos.json and "
                    "eigenvalues.pdf.",
        parents=dirs_parsers + [perfect_band_edge_parser],
        formatter_class=argparse.ArgumentDefaultsHelpFormatter,
        aliases=['beoi'])
    parser_band_edge_orb_infos.add_argument(
        "-y", "--y_range", nargs=2, type=float,
        help="Energy range in y-axis for eigenvalue.pdf")

    parser_band_edge_orb_infos.set_defaults(
        func=make_band_edge_orb_infos_and_eigval_plot)

    # ------------------------------------------------------------------------
    return parser.parse_args(args)


def main():
    args = parse_args_main_vasp(sys.argv[1:])
    args.func(args)


if __name__ == "__main__":
    main()





# -*- coding: utf-8 -*-
#  Copyright (c) 2020. Distributed under the terms of the MIT License.
from pydefect.analyzer.calc_results import CalcResults
from pymatgen.io.vasp import Vasprun, Outcar


def make_calc_results_from_vasp(vasprun: Vasprun,
                                outcar: Outcar) -> CalcResults:
    return CalcResults(structure=vasprun.final_structure,
                       energy=outcar.final_energy,
                       magnetization=outcar.total_mag or 0.0,
                       potentials=[-p for p in outcar.electrostatic_potential],
                       electronic_conv=vasprun.converged_electronic,
                       ionic_conv=vasprun.converged_ionic)



# -*- coding: utf-8 -*-
#  Copyright (c) 2020. Distributed under the terms of the MIT License.
from collections import OrderedDict
from dataclasses import dataclass, asdict
from pathlib import Path
from typing import List

import numpy as np
from monty.json import MSONable
from monty.serialization import loadfn
from ruamel.yaml import add_constructor, resolver, YAML


# courtesy of https://qiita.com/konomochi/items/f5f53ba8efa07ec5089b
add_constructor(resolver.BaseResolver.DEFAULT_MAPPING_TAG,
                lambda loader, node: OrderedDict(loader.construct_pairs(node)))

yaml = YAML()
yaml.default_flow_style = False


@dataclass
class Unitcell(MSONable):
    system: str
    vbm: float
    cbm: float
    ele_dielectric_const: List[List[float]]
    ion_dielectric_const: List[List[float]]

    @property
    def dielectric_constant(self):
        total = (np.array(self.ele_dielectric_const)
                 + np.array(self.ion_dielectric_const))
        return total.tolist()

    def to_yaml_file(self, filename: str = "unitcell.yaml"):
        with open(Path(filename), 'w', encoding='utf-8') as f:
            yaml.dump(asdict(self), f)

    @classmethod
    def from_yaml(cls, filename: str = "unitcell.yaml"):
        return cls(**loadfn(filename))



# -*- coding: utf-8 -*-
#  Copyright (c) 2020 Kumagai group.
from pathlib import Path

import pytest
from pydefect.cli.vasp.get_defect_charge_state import \
    get_defect_charge_state
from pymatgen.core import Structure, Lattice
from pymatgen.io.vasp import Poscar, Potcar, Incar


def test_get_defect_charge_state(tmpdir):
    tmpdir.chdir()
    structure = Structure(Lattice.cubic(10), ["H", "He"], [[0.0]*3, [0.5]*3])
    fake_potcar_str = """ PAW_PBE H 15Jun2001                    
   1.00000000000000     
 parameters from PSCTR are:
   VRHFIN =H: ultrasoft test
   LEXCH  = PE
   TITEL  = PAW_PBE H 15Jun2001
   POMASS =    1.000; ZVAL   =    1.000    mass and valenz
   ENMAX  =  250.000; ENMIN  =  200.000 eV

   END of PSCTR-controll parameters
 End of Dataset
      PAW_PBE He_pv 15Jun2001                    
   1.00000000000000     
 parameters from PSCTR are:
   VRHFIN =He: ultrasoft test
   LEXCH  = PE
   TITEL  = PAW_PBE He_pv 15Jun2001
   POMASS =    1.000; ZVAL   =    1.000    mass and valenz
   ENMAX  =  250.000; ENMIN  =  200.000 eV

   END of PSCTR-controll parameters
 End of Dataset
 """
    Path("POTCAR").write_text(fake_potcar_str)
    potcar = Potcar.from_file("POTCAR")
    incar = Incar.from_string("NELECT = 1")
    assert get_defect_charge_state(Poscar(structure), potcar, incar) == 2 - 1

    structure = Structure(Lattice.cubic(10), ["H", "He", "H", "H"],
                          [[0.0]*3, [0.2]*3, [0.4]*3, [0.6]*3])
    with pytest.raises(ValueError):
        get_defect_charge_state(Poscar(structure), potcar, incar)

    incar = Incar.from_string("")
    assert get_defect_charge_state(Poscar(structure), potcar, incar) == 0





#!/usr/bin/env python
# coding: utf-8

import crystal_toolkit.components as ctc
import dash_html_components as html
from crystal_toolkit.settings import SETTINGS
from jupyter_dash import JupyterDash
from pydefect.analyzer.dash_components.cpd_energy_dash import CpdEnergyComponent
from pydefect.analyzer.defect_energy import DefectEnergySummary, DefectEnergies, \
    DefectEnergy
from pydefect.chem_pot_diag.chem_pot_diag import RelativeEnergies, \
    ChemPotDiagMaker


if __name__ == "__main__":
    app = JupyterDash(suppress_callback_exceptions=True,
                      assets_folder=SETTINGS.ASSETS_PATH)

#    relative_energies = RelativeEnergies({"MgCaO2": -100.0})
    relative_energies = RelativeEnergies({"MgCaBaO2": -100.0})
    # cpd = ChemPotDiagMaker(relative_energies,
    #                        elements=["Mg", "Ca", "O"],
    #                        target="MgCaO2").chem_pot_diag
    cpd = ChemPotDiagMaker(relative_energies,
                           elements=["Mg", "Ca", "Ba", "O"],
                           target="MgCaBaO2").chem_pot_diag

    energy = DefectEnergy(0.0, {"corr": 1.0}, is_shallow=False)
    defect_energy_summary = DefectEnergySummary(
        title="test",
        defect_energies={"Va_O1": DefectEnergies(atom_io={"O": -1},
                                                 charges=[0],
                                                 defect_energies=[energy])},
        rel_chem_pots=cpd.to_target_vertices.chem_pots,
        cbm=2.0, supercell_vbm=-0.2, supercell_cbm=2.2 )

    cpd_e_component = CpdEnergyComponent(cpd, defect_energy_summary)

    my_layout = html.Div([cpd_e_component.layout])
    ctc.register_crystal_toolkit(app=app, layout=my_layout, cache=None)

    app.run_server(port=8099)




# -*- coding: utf-8 -*-
#  Copyright (c) 2020. Distributed under the terms of the MIT License.

from copy import deepcopy
from functools import reduce
from itertools import product
from typing import List, Tuple

import numpy as np
from numpy.linalg import det
from pydefect.util.error_classes import SupercellError
from pymatgen.core import IStructure


class Supercell:
    def __init__(self, input_structure: IStructure, matrix: List[List[int]]):
        self.matrix = matrix
        self.structure = input_structure * matrix
        self.lattice = self.structure.lattice

    @property
    def isotropy(self):
        lengths = self.structure.lattice.lengths
        average = np.average(lengths)
        return sum([abs(length - average) for length in lengths]) / 3 / average

    @property
    def average_angle(self):
        return sum(self.structure.lattice.angles) / 3


class Supercells:
    def __init__(self,
                 input_structure: IStructure,
                 min_num_atoms: int = 50,
                 max_num_atoms: int = 250):
        self.input_structure = input_structure
        min_det = min_num_atoms / len(input_structure) - 1e-5
        max_det = max_num_atoms / len(input_structure) + 1e-5

        self.supercells = []
        matrix = np.eye(3, dtype=int)
        for i in range(50):
            if det(matrix) > max_det:
                break
            if det(matrix) > min_det:
                self.supercells.append(Supercell(input_structure, matrix))
            matrix = self.incremented_matrix(matrix)

    def incremented_matrix(self, matrix: np.ndarray):
        abc = (self.input_structure * matrix).lattice.lengths
        min_indices = [i for i, val in enumerate(abc) if val == min(abc)]
        new_matrix = deepcopy(matrix)
        for index in min_indices:
            new_matrix[index][index] += 1
        return new_matrix

    @property
    def most_isotropic_supercell(self):
        try:
            return min(self.supercells, key=lambda s: s.isotropy)
        except ValueError:
            raise SupercellError


class RhombohedralSupercells(Supercells):
    def incremented_matrix(self, matrix: np.ndarray):
        alpha = (self.input_structure * matrix).lattice.angles[0]

        fan_out = np.array([[1,  1, -1], [-1,  1,  1], [1, -1,  1]])
        close_in = np.array([[1, 0, 1], [1, 1, 0], [0, 1, 1]])

        if alpha < 90:
            return np.dot(matrix, fan_out)
        else:
            return np.dot(matrix, close_in)


class TetragonalSupercells(Supercells):
    def incremented_matrix(self, matrix: np.ndarray):
        a, _, c = (self.input_structure * matrix).lattice.lengths
        new_matrix = deepcopy(matrix)
        if a > c:
            new_matrix[2][2] += 1
            return new_matrix
        else:
            ab_submatrix_det = round(det(matrix[0:2, 0:2]))
            x, y = self.next_x_y_combination(ab_submatrix_det)
            new_matrix[0:2, 0:2] = self.matrix_from_x_y(x, y)
            return new_matrix

    @staticmethod
    def matrix_from_x_y(x: int, y: int) -> np.ndarray:
        """ """
        assert x >= 1
        assert y >= 0

        expand_matrix = np.array([[x, 0], [0, x]])
        if y == 0:
            rot_matrix = np.eye(2)
        elif y == 1:
            rot_matrix = np.array([[1, 1], [-1, 1]])
        else:
            matrix_list = [np.array([[1, 1], [-1, 1]])] * y
            rot_matrix = reduce(np.dot, matrix_list)

        return np.dot(expand_matrix, rot_matrix)

    @staticmethod
    def next_x_y_combination(num: int) -> Tuple[int, int]:
        """ Find x, y with minimum of x ** 2 * 2^y larger than num."""
        pairs = {(x, y): x * x * 2 ** y
                 for x, y in product(range(1, 10), range(8))}
        filtered = {k: v for k, v in pairs.items() if v > num}
        return next(iter(sorted(filtered, key=lambda u: pairs[u])))





# -*- coding: utf-8 -*-
#  Copyright (c) 2020. Distributed under the terms of the MIT License.
from collections.abc import Set
from pathlib import Path
from typing import Iterator, List, Set as typeSet

import yaml
from monty.serialization import loadfn
from pydefect.input_maker.defect import screen_simple_defect, SimpleDefect


class DefectSet(Set):
    def __init__(self, defects: typeSet[SimpleDefect]):
        self.defects = defects

    def __contains__(self, defect: object) -> bool:
        return defect in self.defects

    def __len__(self) -> int:
        return len(self.defects)

    def __iter__(self) -> Iterator:
        yield from self.defects

    def to_yaml(self, filename: str = "defect_in.yaml") -> None:
        d = {defect.name: list(defect.charges) for defect in self.defects}
        Path(filename).write_text(yaml.dump(d, default_flow_style=None))

    @classmethod
    def from_yaml(cls, filename: str = "defect_in.yaml") -> "DefectSet":
        d = loadfn(filename)
        names = set()
        for name, charges in d.items():
            in_name, out_name = name.split("_")
            names.add(SimpleDefect(in_name, out_name, tuple(charges)))
        return cls(names)


def screen_defect_set(defect_set: DefectSet, keywords: List[str]):
    result = set()
    for defect in defect_set:
        screened = screen_simple_defect(defect, keywords)
        if screened:
            result.add(screened)
    return DefectSet(result)



# -*- coding: utf-8 -*-
#  Copyright (c) 2020. Distributed under the terms of the MIT License.
from dataclasses import dataclass
from pathlib import Path
from typing import Tuple, Optional

import numpy as np
import yaml
from monty.json import MSONable, MontyDecoder
from pydefect.analyzer.defect_structure_comparator import \
    DefectStructureComparator
from pydefect.util.coords import pretty_coords
from pymatgen.core import IStructure
from vise.util.mix_in import ToJsonFileMixIn
from vise.util.structure_symmetrizer import StructureSymmetrizer
from vise.util.typing import Coords


@dataclass(frozen=True)
class PerturbedSite(MSONable):
    element: str
    distance: float
    initial_coords: Coords
    perturbed_coords: Coords
    displacement: float

    @classmethod
    def from_dict(cls, d):
        return cls(element=d["element"],
                   distance=d["distance"],
                   initial_coords=tuple(d["initial_coords"]),
                   perturbed_coords=tuple(d["perturbed_coords"]),
                   displacement=d["displacement"])

    def __str__(self):
        elem = f"{self.element:>4}"
        dist = f"{self.distance:4.2f}"
        ini_coords = pretty_coords(self.initial_coords)
        p_coords = pretty_coords(self.perturbed_coords)
        disp = f"{self.displacement:7.2f}"
        return f"{elem} {dist} {ini_coords} -> {p_coords} {disp}"


@dataclass(frozen=True)
class DefectEntry(MSONable, ToJsonFileMixIn):
    name: str
    charge: int
    structure: IStructure
    site_symmetry: str
    defect_center: Coords
    perturbed_structure: Optional[IStructure] = None
    perturbed_sites: Optional[Tuple[PerturbedSite, ...]] = None
    perturbed_site_symmetry: Optional[str] = None

    @classmethod
    def from_dict(cls, d):
        d = {k: MontyDecoder().process_decoded(v) for k, v in d.items() if not k.startswith("@")}
        if d["perturbed_sites"]:
            d["perturbed_sites"] = tuple([MontyDecoder().process_decoded(d) for d in d["perturbed_sites"]])
        d["defect_center"] = tuple(d["defect_center"])
        return cls(**d)

    @property
    def anchor_atom_index(self) -> int:
        """ Returns an index of atom that is the farthest from the defect.

         Only the first occurrence is returned when using argmax.
         docs.scipy.org/doc/numpy-1.15.0/reference/generated/numpy.argmax.html
         """
        lattice = self.structure.lattice
        all_coords = self.structure.frac_coords
        dist_set = lattice.get_all_distances(self.defect_center, all_coords)[0]
        return int(np.argmax(dist_set))

    @property
    def anchor_atom_coords(self) -> np.ndarray:
        return self.structure[self.anchor_atom_index].frac_coords

    @property
    def perturbed_site_indices(self):
        result = []
        for i, (s1, s2) in enumerate(zip(self.structure,
                                         self.perturbed_structure)):
            if (s1.frac_coords != s2.frac_coords).any():
                result.append(i)
        return result

    @property
    def full_name(self):
        return "_".join([self.name, str(self.charge)])

    def to_prior_info(self, filename):
        d = {"charge": self.charge}
        Path(filename).write_text(yaml.dump(d))

    def __str__(self):
        result = f""" -- defect entry info
name: {self.full_name}
site symmetry: {self.site_symmetry}
defect center: ({", ".join([f"{c:6.3f}" for c in self.defect_center])})
perturbed sites:
elem dist   initial_coords             perturbed_coords         displacement
"""
        if self.perturbed_sites:
            result += "\n".join([i.__str__() for i in self.perturbed_sites])

        return result


def make_defect_entry(name: str,
                      charge: int,
                      perfect_structure: IStructure,
                      defect_structure: IStructure):

    analyzer = DefectStructureComparator(perfect_structure, defect_structure)

    species = []
    frac_coords = []
    for d, p in enumerate(analyzer.p_to_d):
        if p is None:
            site = defect_structure[d]
        else:
            site = perfect_structure[p]
        species.append(site.specie)
        frac_coords.append(site.frac_coords)

    initial_structure = IStructure(perfect_structure.lattice,
                                   species, frac_coords)
    symmetrizer = StructureSymmetrizer(initial_structure)

    return DefectEntry(name=name,
                       charge=charge,
                       structure=initial_structure,
                       site_symmetry=symmetrizer.point_group,
                       defect_center=tuple(analyzer.defect_center_coord))



# -*- coding: utf-8 -*-
#  Copyright (c) 2020. Distributed under the terms of the MIT License.
from pathlib import Path

from monty.serialization import loadfn
from pymatgen.core import Structure, Lattice

from pydefect.cli.vasp.make_poscars_from_query import make_poscars_from_query


def test_make_poscars_from_query(tmpdir):
    mg_structure = Structure(Lattice.cubic(1), species=["Mg", "Mg"],
                             coords=[[0, 0, 0], [0.5, 0.5, 0.5]])

    query = [{"full_formula": "Mg2", "task_id": "mp-1",
              "total_magnetization": 0.1,
              "band_gap": 0.0,
              "structure": mg_structure},
             {"full_formula": "O16"}]
    make_poscars_from_query(query, path=Path(tmpdir))
    tmpdir.chdir()
    s_mg = Structure.from_file(Path("Mg_mp-1") / "POSCAR")
    assert s_mg == mg_structure

    actual = loadfn(Path("Mg_mp-1") / "prior_info.yaml")
    expected = {'band_gap': 0.0, 'data_source': 'mp-1', 'total_magnetization': 0.1}
    assert actual == expected

    actual = Structure.from_file(Path("mol_O2") / "POSCAR")
    expected = Structure.from_str("""Default POSCAR of O2 molecule
   1.00000000000000     
    10.0000000000000000    0.0000000000000000    0.0000000000000000
     0.0000000000000000   10.0000000000000000    0.0000000000000000
     0.0000000000000000    0.0000000000000000   10.0000000000000000
   O 
     2
Direct
  0.0145089778962415  0.0000000000000000  0.0000000000000000
  0.1354910221037571  0.0000000000000000  0.0000000000000000""", fmt="poscar")
    assert actual == expected




# -*- coding: utf-8 -*-
#  Copyright (c) 2020. Distributed under the terms of the MIT License.

import pytest
from pydefect.analyzer.band_edge_states import BandEdgeOrbitalInfos, OrbitalInfo
from pydefect.analyzer.eigenvalue_plotter import EigenvalueMplPlotter

try:
    import psutil
    PSUTIL_NOT_PRESENT = False
except ModuleNotFoundError:
    PSUTIL_NOT_PRESENT = True

orb = [[OrbitalInfo(0.0, orbitals={}, occupation=1.0, participation_ratio=0.3),
        OrbitalInfo(0.5, orbitals={}, occupation=0.5, participation_ratio=0.3),
        OrbitalInfo(1.0, orbitals={}, occupation=0.0, participation_ratio=0.3)],
       [OrbitalInfo(0.0, orbitals={}, occupation=1.0, participation_ratio=0.3),
        OrbitalInfo(0.5, orbitals={}, occupation=0.5, participation_ratio=0.3),
        OrbitalInfo(1.0, orbitals={}, occupation=0.0, participation_ratio=0.3)]]

orbital_infos = BandEdgeOrbitalInfos(
    orbital_infos=[orb, orb],
    kpt_coords=[(0.0, 0.0, 0.0), (0.25, 0.0, 0.0)],
    kpt_weights=[0.5, 0.5],
    lowest_band_index=10,
    fermi_level=0.5)


@pytest.fixture
def eigenvalue_mpl_plotter():
    return EigenvalueMplPlotter(title="test",
                                band_edge_orb_infos=orbital_infos,
                                supercell_vbm=0.1, supercell_cbm=0.9)


def test_plot(eigenvalue_mpl_plotter):
    eigenvalue_mpl_plotter.construct_plot()
    eigenvalue_mpl_plotter.plt.show()


def test_plot_wo_spin():
    be_orbital_info = BandEdgeOrbitalInfos(
        orbital_infos=[orb],
        kpt_coords=[(0.0, 0.0, 0.0), (0.25, 0.0, 0.0)],
        kpt_weights=[0.5, 0.5],
        lowest_band_index=10,
        fermi_level=0.5)
    plotter = EigenvalueMplPlotter(title="test",
                                   band_edge_orb_infos=be_orbital_info,
                                   supercell_vbm=0.1, supercell_cbm=0.9)
    plotter.construct_plot()
    plotter.plt.show()

# def test_plot_with_actual_file():
#     eig = loadfn(Path(__file__).parent / "band_edge_eigenvalues.json")
#     plotter = EigenvalueMplPlotter(title="test", band_edge_eigenvalues=eig,
#                                    supercell_vbm=0.5, supercell_cbm=2.6)
#     plotter.construct_plot()
#     plotter.plt.show()


# @pytest.mark.skipif(PSUTIL_NOT_PRESENT, reason="psutil does not exist")
# def test_plotly_with_actual_file():
#     eig = loadfn(Path(__file__).parent / "band_edge_eigenvalues.json")
#     plotter = EigenvaluePlotlyPlotter(title="test", band_edge_eigenvalues=eig,
#                                       supercell_vbm=0.5, supercell_cbm=2.6)
#     fig = plotter.create_figure()
#     # fig.show()
#     show_png(fig)



# -*- coding: utf-8 -*-
#  Copyright (c) 2020 Kumagai group.

import pytest
from pydefect.chem_pot_diag.chem_pot_diag import CompositionEnergy, \
    CompositionEnergies
from pydefect.cli.vasp.make_composition_energies_from_mp import \
    make_composition_energies_from_mp, remove_higher_energy_comp
from pymatgen.core import Composition, Element
from pymatgen.entries.computed_entries import ComputedEntry


@pytest.fixture
def mp_query_mock(mocker):
    mock = mocker.patch(
        "pydefect.cli.vasp.make_composition_energies_from_mp.MpEntries")
    mock.return_value.materials = \
        [ComputedEntry(Composition("O8"), -39.58364375, entry_id="mp-1"),
         ComputedEntry(Composition("Mg3"), -4.79068775, entry_id="mp-2"),
         ComputedEntry(Composition("Mg1O1"), -11.96742144, entry_id="mp-3")]
    return mock


@pytest.fixture
def comp_energies():
    return CompositionEnergies({
        Composition('O8'): CompositionEnergy(-39.58364375, "mp-1"),
        Composition('Mg3'): CompositionEnergy(-4.79068775, "mp-2"),
        Composition('Mg1O1'): CompositionEnergy(-11.96742144, "mp-3")})


# mp
# O:  -1.54794890
# Mg: -0.01731593
diff = {"Mg": -0.00912097 - -0.01731593, "O": -1.61154565 - -1.54794890}


@pytest.fixture
def comp_energies_corr():
    return CompositionEnergies({
        Composition('O8'): CompositionEnergy(-39.58364375 + diff["O"] * 8, "mp-1"),
        Composition('Mg3'): CompositionEnergy(-4.79068775 + diff["Mg"] * 3, "mp-2"),
        Composition('Mg1O1'): CompositionEnergy(-11.96742144 + diff["Mg"] + diff["O"], "mp-3")})


def test_make_chem_pot_diag_from_mp(mp_query_mock, comp_energies):
    actual = make_composition_energies_from_mp(elements=["Mg", "O"])
    mp_query_mock.assert_called_once_with(["Mg", "O"])
    assert actual == comp_energies


def test_make_chem_pot_diag_from_mp_yaml(mp_query_mock, tmpdir,
                                         comp_energies_corr):
    tmpdir.chdir()
    tmpdir.join("tmp.yaml").write("""O:  -1.61154565
Mg: -0.00912097""")
    actual = make_composition_energies_from_mp(elements=["Mg", "O"],
                                               atom_energy_yaml="tmp.yaml")
    assert actual == comp_energies_corr


def test_remove_higher_energy_comp():
    comp_es = CompositionEnergies(
        {Composition('Mg1'): CompositionEnergy(-1.9, "mp-2"),
         Composition('Mg2'): CompositionEnergy(-4.0, "mp-3"),
         Composition('O2'): CompositionEnergy(-4.0, "mp-4")})
    actual = remove_higher_energy_comp(comp_es)
    expected = CompositionEnergies(
        {Composition('Mg2'): CompositionEnergy(-4.0, "mp-3"),
         Composition('O2'): CompositionEnergy(-4.0, "mp-4"),})

    assert actual == expected




# -*- coding: utf-8 -*-
#  Copyright (c) 2020. Distributed under the terms of the MIT License.
from dataclasses import dataclass
from typing import List, Optional

from monty.json import MSONable
from pymatgen.core import IStructure
from vise.util.logger import get_logger
from vise.util.mix_in import ToJsonFileMixIn


logger = get_logger(__name__)


@dataclass
class CalcResults(MSONable, ToJsonFileMixIn):
    # keep structure and site_symmetry in CalcResults
    structure: IStructure
    energy: float
    magnetization: float
    # potential acting on the positive unit charge, whose sign is reserved from
    # vasp convention of atomic site potential.
    potentials: List[float]
    electronic_conv: Optional[bool] = None
    ionic_conv: Optional[bool] = None

    def show_convergence_warning(self):
        if self.electronic_conv is False:
            logger.warning("SCF is not converged.")
        if self.ionic_conv is False:
            logger.warning("Ionic relaxation is not converged.")

    def __str__(self):
        return f""" -- calc results info
energy: {self.energy:10.3f}
magnetization: {self.magnetization:6.2f}
electronic convergence: {self.electronic_conv}
ionic convergence: {self.ionic_conv}"""


class NoElectronicConvError(AssertionError):
    pass


class NoIonicConvError(AssertionError):
    pass



# -*- coding: utf-8 -*-
#  Copyright (c) 2020. Distributed under the terms of the MIT License.
import pytest
from pydefect.analyzer.calc_results import CalcResults
from pymatgen.core import IStructure, Lattice

from vise.tests.helpers.assertion import assert_json_roundtrip


@pytest.fixture
def calc_results():
    return CalcResults(structure=IStructure(Lattice.cubic(1.0), ["H"], [[0.0]*3]),
                       energy=1.0,
                       magnetization=0.0,
                       potentials=[0.0],
                       electronic_conv=False,
                       ionic_conv=False)


def test_calc_results_json_roundtrip(calc_results, tmpdir):
    assert_json_roundtrip(calc_results, tmpdir)


def test_calc_results_show_convergence_warning(calc_results):
    calc_results.show_convergence_warning()


def test_calc_results_str(calc_results):
    assert calc_results.__str__() == """ -- calc results info
energy:      1.000
magnetization:   0.00
electronic convergence: False
ionic convergence: False"""



# -*- coding: utf-8 -*-
#  Copyright (c) 2020. Distributed under the terms of the MIT License.
from dataclasses import dataclass, field
from typing import List, Dict

from monty.json import MSONable
from numpy.linalg import det
from pydefect.database.database import electronegativity, oxidation_state
from pydefect.util.structure_tools import Distances
from pymatgen.core import IStructure
from vise.util.logger import get_logger
from vise.util.mix_in import ToJsonFileMixIn
from vise.util.structure_symmetrizer import Site

logger = get_logger(__name__)


@dataclass(frozen=True)
class Interstitial(MSONable):
    frac_coords: List[float]
    site_symmetry: str
    info: str = None


@dataclass(frozen=True)
class SupercellInfo(MSONable, ToJsonFileMixIn):
    structure: IStructure
    space_group: str
    transformation_matrix: List[List[int]]
    sites: Dict[str, Site]
    interstitials: List[Interstitial] = field(default_factory=list)
    unitcell_structure: IStructure = None

    @classmethod
    def from_dict(cls, d):
        if "transform_matrix" in d:
            d["transformation_matrix"] = d.pop("transform_matrix")
        return super().from_dict(d)

    def coords(self, name):
        site = self.sites[name]
        coord = list(self.structure[site.equivalent_atoms[0]].frac_coords)
        distances = Distances(self.structure, coord)
        return distances.coordination()

    def interstitial_coords(self, idx: int):
        interstitial = self.interstitials[idx]
        distances = Distances(self.structure, interstitial.frac_coords)
        return distances.coordination(include_on_site=True)

    @property
    def multiplicity(self):
        return int(round(det(self.transformation_matrix)))

    @property
    def _transform_matrix_str(self):
        return '  '.join(str(x) for x in self.transformation_matrix)

    def _frac_coords(self, site):
        repr_atom_idx = site.equivalent_atoms[0]
        frac_coords = self.structure[repr_atom_idx].frac_coords
        return "{0[0]:9.7f}  {0[1]:9.7f}  {0[2]:9.7f}".format(frac_coords)

    @staticmethod
    def _stripe_numbers(name):
        return ''.join([i for i in name if i.isalpha()])

    def __str__(self):
        lines = [f"Space group: {self.space_group}",
                 f"Transformation matrix: {self._transform_matrix_str}",
                 f"Cell multiplicity: {self.multiplicity}", ""]

        for name, site in self.sites.items():
            elem = self._stripe_numbers(name)
            coordination = self.coords(name).distance_dict
            lines.append(f"   Irreducible element: {name}")
            lines.append(f"        Wyckoff letter: {site.wyckoff_letter}")
            lines.append(f"         Site symmetry: {site.site_symmetry}")
            lines.append(f"         Cutoff radius: {self.coords(name).cutoff}")
            lines.append(f"          Coordination: {coordination}")
            lines.append(f"      Equivalent atoms: {site.pprint_equiv_atoms}")
            lines.append(f"Fractional coordinates: {self._frac_coords(site)}")
            lines.append(f"     Electronegativity: {electronegativity(elem)}")
            lines.append(f"       Oxidation state: {oxidation_state(elem)}")
            lines.append("")
        lines.append("-- interstitials")

        for idx, site in enumerate(self.interstitials):
            coordination = self.interstitial_coords(idx).distance_dict

            lines.append(f"#{idx + 1}")
            if site.info:
                lines.append(f"                  Info: {site.info}")

            frac = \
                "{0[0]:9.7f}  {0[1]:9.7f}  {0[2]:9.7f}".format(site.frac_coords)
            lines.append(f"Fractional coordinates: {frac}")
            lines.append(f"         Site symmetry: {site.site_symmetry}")
            lines.append(f"          Coordination: {coordination}")
            lines.append("")

        return "\n".join(lines)



# -*- coding: utf-8 -*-
#  Copyright (c) 2020. Distributed under the terms of the MIT License.

import numpy as np
from pydefect.analyzer.calc_results import CalcResults
from pydefect.cli.vasp.make_efnv_correction import \
    calc_max_sphere_radius, make_efnv_correction
from pydefect.corrections.efnv_correction import PotentialSite
from pymatgen.core import IStructure, Lattice


def test_make_efnv_correction(mocker):
    mock_perfect = mocker.Mock(spec=CalcResults, autospec=True)
    mock_defect = mocker.Mock(spec=CalcResults, autospec=True)

    mock_perfect.structure = IStructure(
        Lattice.cubic(10), species=["H"] + ["He"] * 3 + ["Li"],
        coords=[[0, 0, 0], [1/2, 1/2, 0], [1/2, 0, 1/2], [0, 1/2, 1/2], [0, 0, 1/2]])
    mock_defect.structure = IStructure(
        Lattice.cubic(10), species=["He"] * 3 + ["Li"],
        coords=[[1/2, 1/2, 0], [1/2, 0, 1/2], [0, 1/2, 1/2], [0, 0, 1/2]])
    mock_perfect.potentials = [3.0, 4.0, 5.0, 6.0, 7.0]
    mock_defect.potentials = [14.0, 25.0, 36.0, 47.0]

    mock_ewald = mocker.patch("pydefect.cli.vasp.make_efnv_correction.Ewald")
    ewald = mocker.Mock()
    ewald.lattice_energy = 1e3
    ewald.atomic_site_potential.return_value = 1e4
    mock_ewald.return_value = ewald

    efnvc = make_efnv_correction(charge=2,
                                 calc_results=mock_defect,
                                 perfect_calc_results=mock_perfect,
                                 dielectric_tensor=np.eye(3))

    unit_conversion = 180.95128169876497

    assert efnvc.charge == 2
    assert efnvc.point_charge_correction == -4e3 * unit_conversion
    assert efnvc.defect_region_radius == 5.0
    assert efnvc.sites == [PotentialSite("He", 5 * np.sqrt(2), 10.0, 2e4 * unit_conversion),
                           PotentialSite("He", 5 * np.sqrt(2), 20.0, 2e4 * unit_conversion),
                           PotentialSite("He", 5 * np.sqrt(2), 30.0, 2e4 * unit_conversion),
                           PotentialSite("Li", 5.0, 40.0, None),
                           ]


def test_calc_max_sphere_radius():
    lattice_vectors_1 = np.array([[5, 0, 0], [0, 10, 0], [0, 0, 20]])
    lattice_vectors_2 = np.array([[10, 0, 0], [0, 10, 0], [10, 10, 10]])
    radius_1 = calc_max_sphere_radius(lattice_vectors_1)
    radius_2 = calc_max_sphere_radius(lattice_vectors_2)
    assert radius_1 == 10.0
    assert radius_2 == 5.0




# -*- coding: utf-8 -*-
#  Copyright (c) 2020 Kumagai group.
from pathlib import Path

from monty.serialization import loadfn
from pydefect.analyzer.calc_results import CalcResults
from pydefect.cli.main_functions import get_calc_results
from pydefect.cli.main_tools import parse_dirs
from pydefect.cli.make_defect_vesta_file import MakeDefectVestaFile
from pydefect.cli.vasp.make_composition_energies_from_mp import \
    make_composition_energies_from_mp
from pydefect.cli.vasp.make_gkfo_correction import make_gkfo_correction
from pydefect.corrections.site_potential_plotter import SitePotentialMplPlotter
from vise.util.logger import get_logger


logger = get_logger(__name__)


def composition_energies_from_mp(args) -> None:
    composition_energies = make_composition_energies_from_mp(
        elements=args.elements, atom_energy_yaml=args.atom_energy_yaml)
    composition_energies.to_yaml_file()


def add_interstitials_from_local_extrema(args) -> None:
    supercell_info = args.local_extrema.append_sites_to_supercell_info(
        args.supercell_info, args.indices)
    supercell_info.to_json_file()


def make_defect_vesta_file(args) -> None:
    def _inner(_dir: Path):
        defect_str_info = loadfn(_dir / "defect_structure_info.json")
        make_vesta_file = MakeDefectVestaFile(defect_str_info,
                                              args.cutoff,
                                              args.min_displace_w_arrows,
                                              args.arrow_factor,
                                              args.title)
        make_vesta_file.initial_vesta.write_file(_dir / "defect_initial.vesta")
        make_vesta_file.final_vesta.write_file(_dir / "defect.vesta")

    parse_dirs(args.dirs, _inner)


def make_gkfo_correction_from_vasp(args):
    gkfo = make_gkfo_correction(
        efnv_correction=args.initial_efnv_correction,
        additional_charge=args.charge_diff,
        final_calc_results=args.final_calc_results,
        initial_calc_results=args.initial_calc_results,
        diele_tensor=args.unitcell.dielectric_constant,
        ion_clamped_diele_tensor=args.unitcell.ele_dielectric_const)
    print(gkfo)
    gkfo.to_json_file("gkfo_correction.json")
    plotter = SitePotentialMplPlotter.from_gkfo_corr(
        title="GKFO correction", gkfo_correction=gkfo)
    plotter.construct_plot()
    plotter.plt.savefig(fname="gkfo_correction.pdf")
    plotter.plt.clf()





# -*- coding: utf-8 -*-
#  Copyright (c) 2020 Kumagai group.
from dataclasses import dataclass
from itertools import product
from typing import Tuple, List

import numpy as np
from pydefect.cli.vasp.make_efnv_correction import calc_max_sphere_radius
from pymatgen.core import Lattice
from pymatgen.io.vasp import Chgcar


@dataclass
class Grids:
    lattice: Lattice
    dim: Tuple[int, int, int]
    distance_data: np.ndarray

    def dump(self, filename="grids.npz"):
        np.savez(filename, matrix=self.lattice.matrix,
                 distance_data=self.distance_data)

    @classmethod
    def from_file(cls, filename="grids.npz"):
        loaded_dict = np.load(filename)
        lattice = Lattice(loaded_dict["matrix"])
        return cls(dim=loaded_dict["distance_data"].shape, lattice=lattice,
                   distance_data=loaded_dict["distance_data"])

    @classmethod
    def from_chgcar(cls, chgcar: Chgcar):
        lattice, dim = chgcar.structure.lattice, chgcar.dim
        grid_points = [[x / dim[0], y / dim[1], z / dim[2]]
                       for (x, y, z) in product(*[list(range(i)) for i in dim])]
        # Use boolean indexing to find charges within the desired distance.
        # data[:, 0]: shifted_coords
        # data[:, 1]: distances
        # data[:, 2]: sequential indices
        # data[:, 3]: images
        distances_data = np.array(lattice.get_all_distances(
            grid_points, (0.0, 0.0, 0.0)))
        return cls(lattice, dim, distances_data[:, 0].reshape(dim))

    def shifted_distance_data(self, center: List[int]):
        return np.roll(np.roll(np.roll(self.distance_data, center[0], axis=0),
                               center[1], axis=1), center[2], axis=2)

    def spherical_dist(self,
                       data: np.ndarray,
                       center: List[int],
                       distance_bins: np.ndarray):
        assert distance_bins[-1] <= calc_max_sphere_radius(self.lattice.matrix)
        shifted_dist = self.shifted_distance_data(center)
        _sum, _ = np.histogram(shifted_dist, distance_bins, weights=data)
        counts, _ = np.histogram(shifted_dist, distance_bins)
        histogram = _sum / counts / self.lattice.volume
        return histogram.tolist()

#        num_bins = int(np.ceil(radius / bin_interval))
#        distance_bins = [bin_interval * i for i in range(num_bins)] + [radius]
    # distance_bins: List[float]



# -*- coding: utf-8 -*-
#  Copyright (c) 2020. Distributed under the terms of the MIT License.
from dataclasses import dataclass
from typing import List, Dict, Tuple, Optional, Set

import numpy as np
from monty.json import MSONable
from pydefect.defaults import defaults
from pydefect.util.coords import pretty_coords
from tabulate import tabulate
from vise.util.mix_in import ToJsonFileMixIn
from vise.util.typing import Coords, GenCoords

printed_orbital_weight_threshold = 0.1


@dataclass
class BandEdgeEigenvalues(MSONable, ToJsonFileMixIn):
    # [spin, k-idx, band-idx] = energy, occupation
    energies_and_occupations: List[List[List[List[float]]]]
    kpt_coords: List[List[float]]
    lowest_band_index: int


def pretty_orbital(orbitals: Dict[str, List[float]]):
    """
    :param orbitals: An example is {"Mn": [0.5, 0.4, 0.01]} (no f-orbital)
    :return: "Mn-s: 0.50, Mn-p: 0.40"
    """
    orbital_infos = []
    for elem, orbs in orbitals.items():
        for orb_name, weight in zip(["s", "p", "d", "f"], orbs):
            if weight > printed_orbital_weight_threshold:
                orbital_infos.append(f"{elem}-{orb_name}: {weight:.2f}")
    return ", ".join(orbital_infos)


@dataclass
class OrbitalInfo(MSONable):
    """Note that this is the code- and version-dependent quantities. """
    energy: float  # max eigenvalue
    # {"Mn": [0.01, ..], "O": [0.03, 0.5]},
    # where lists contain s, p, d, (f) orbital components.
    orbitals: Dict[str, List[float]]
    occupation: float
    participation_ratio: float = None


@dataclass
class BandEdgeOrbitalInfos(MSONable, ToJsonFileMixIn):
    orbital_infos: List[List[List["OrbitalInfo"]]]  # [spin, k-idx, band-idx]
    kpt_coords: List[Coords]
    kpt_weights: List[float]
    lowest_band_index: int
    fermi_level: float

    def kpt_idx(self, kpt_coord):
        for i, orig_kpt in enumerate(self.kpt_coords):
            if sum(abs(k - l) for k, l in zip(orig_kpt, kpt_coord)) < 1e-5:
                return i
        raise ValueError(f"{kpt_coord} not in kpt_coords {self.kpt_coords}.")

    @property
    def energies_and_occupations(self) -> List[List[List[List[float]]]]:
        result = np.zeros(np.shape(self.orbital_infos) + (2,))
        for i, x in enumerate(self.orbital_infos):
            for j, y in enumerate(x):
                for k, z in enumerate(y):
                    result[i][j][k] = [z.energy, z.occupation]
        return result.tolist()

    def __str__(self):
        return "\n".join([" -- band-edge orbitals info",
                          "K-points info",
                          self._kpt_block,
                          "",
                          "Band info near band edges",
                          self._band_block])

    @property
    def _band_block(self):
        band_block = [["Index", "Kpoint index", "Energy", "Occupation",
                       "P-ratio", "Orbital"]]
        for orbital_info in self.orbital_infos:
            max_idx, min_idx, t_orb_info = self._band_idx_range(orbital_info)
            for band_idx in range(min_idx, max_idx):
                actual_band_idx = band_idx + self.lowest_band_index + 1
                for kpt_idx, orb_info in enumerate(t_orb_info[band_idx], 1):
                    energy = f"{orb_info.energy :5.2f}"
                    occupation = f"{orb_info.occupation:4.1f}"
                    p_ratio = f"{orb_info.participation_ratio:4.1f}"
                    orbs = pretty_orbital(orb_info.orbitals)
                    band_block.append([actual_band_idx, kpt_idx, energy,
                                       occupation, p_ratio, orbs])
                band_block.append(["--"])
            band_block.append("")
        return tabulate(band_block, tablefmt="plain")

    @property
    def _kpt_block(self):
        kpt_block = [["Index", "Coords", "Weight"]]
        for index, (kpt_coord, kpt_weight) in enumerate(
                zip(self.kpt_coords, self.kpt_weights), 1):
            coord = pretty_coords(kpt_coord)
            weight = f"{kpt_weight:4.3f}"
            kpt_block.append([index, coord, weight])
        return tabulate(kpt_block, tablefmt="plain")

    @staticmethod
    def _band_idx_range(orbital_info: List[List[OrbitalInfo]]
                        ) -> Tuple[int, int, List[List[OrbitalInfo]]]:
        # swap [kpt_idx][band_idx] -> [bane_idx][kpt_idx]
        t_orbital_info = np.array(orbital_info).T.tolist()
        middle_idx = int(len(t_orbital_info) / 2)
        for band_idx, (upper, lower) in enumerate(zip(t_orbital_info[1:],
                                                      t_orbital_info[:-1])):
            # determine the band_idx where the occupation changes largely.
            if lower[0].occupation - upper[0].occupation > 0.1:
                middle_idx = band_idx + 1
                break
        max_idx = min(middle_idx + 3, len(t_orbital_info))
        min_idx = max(middle_idx - 3, 0)
        return max_idx, min_idx, t_orbital_info


@dataclass
class LocalizedOrbital(MSONable):
    band_idx: int
    ave_energy: float
    occupation: float
    orbitals: Dict[str, List[float]]
    participation_ratio: Optional[float] = None
    radius: Optional[float] = None
    center: Optional[GenCoords] = None


@dataclass
class EdgeInfo(MSONable):
    """Information for a particular band edge, namely VBM or CBM."""
    band_idx: int
    kpt_coord: Coords
    orbital_info: "OrbitalInfo"

    @property
    def orbitals(self):
        return self.orbital_info.orbitals

    @property
    def energy(self):
        return self.orbital_info.energy

    @property
    def occupation(self):
        return self.orbital_info.occupation

    @property
    def p_ratio(self):
        return self.orbital_info.participation_ratio


@dataclass
class PerfectBandEdgeState(MSONable, ToJsonFileMixIn):
    vbm_info: EdgeInfo
    cbm_info: EdgeInfo

    def __str__(self):
        def show_edge_info(edge_info: EdgeInfo):
            return [edge_info.band_idx,
                    edge_info.energy,
                    f"{edge_info.occupation:5.2f}",
                    pretty_orbital(edge_info.orbital_info.orbitals),
                    pretty_coords(edge_info.kpt_coord)]

        return tabulate([
            ["", "Index", "Energy", "Occupation", "Orbitals", "K-point coords"],
            ["VBM"] + show_edge_info(self.vbm_info),
            ["CBM"] + show_edge_info(self.cbm_info)], tablefmt="plain")


@dataclass
class BandEdgeState(MSONable):
    """Information related to the band edge in a defect supercell.

    E.g.,
    # 124   --   <- CBM
    # 123   --   <- localized orbital
    # 122   --   <- localized orbital
    # 121   --   <- VBM
    """

    vbm_info: EdgeInfo
    cbm_info: EdgeInfo
    vbm_orbital_diff: float  # Difference from those of perfect supercell.
    cbm_orbital_diff: float
    localized_orbitals: List[LocalizedOrbital]
    vbm_hole_occupation: float = None
    cbm_electron_occupation: float = None

    @property
    def is_shallow(self):
        return self.has_donor_phs or self.has_acceptor_phs

    @property
    def has_donor_phs(self):
        return self.cbm_electron_occupation > defaults.state_occupied_threshold

    @property
    def has_acceptor_phs(self):
        return self.vbm_hole_occupation > defaults.state_occupied_threshold

    @property
    def has_unoccupied_localized_state(self):
        return any([lo.occupation < defaults.state_unoccupied_threshold
                    for lo in self.localized_orbitals])

    @property
    def has_occupied_localized_state(self):
        return any([lo.occupation > defaults.state_occupied_threshold
                    for lo in self.localized_orbitals])

    def __str__(self):
        return "\n".join([self._edge_info,
                          "---", "Localized Orbital(s)",
                          self._orbital_info])

    @property
    def _orbital_info(self):
        inner_table = [["Index", "Energy", "P-ratio", "Occupation", "Orbitals"]]
        w_radius = self.localized_orbitals and self.localized_orbitals[0].radius
        if w_radius:
            inner_table[0].extend(["Radius", "Center"])

        for lo in self.localized_orbitals:
            participation_ratio = f"{lo.participation_ratio:5.2f}" \
                if lo.participation_ratio else "None"
            inner = [lo.band_idx + 1,
                     f"{lo.ave_energy:7.3f}",
                     participation_ratio,
                     f"{lo.occupation:5.2f}",
                     pretty_orbital(lo.orbitals)]
            if w_radius:
                inner.extend([f"{lo.radius:5.2f}", pretty_coords(lo.center)])
            inner_table.append(inner)

        return tabulate(inner_table, tablefmt="plain")

    @property
    def _edge_info(self):
        inner_table = [["", "Index", "Energy", "P-ratio", "Occupation",
                        "OrbDiff", "Orbitals", "K-point coords"],
                       ["VBM"] + self._show_edge_info(
                           self.vbm_info, self.vbm_orbital_diff),
                       ["CBM"] + self._show_edge_info(
                           self.cbm_info, self.cbm_orbital_diff)]
        table = tabulate(inner_table, tablefmt="plain")
        vbm_phs = f"vbm has acceptor phs: {self.has_acceptor_phs} " \
                  f"({self.vbm_hole_occupation:5.3f} vs. {defaults.state_occupied_threshold})"
        cbm_phs = f"cbm has donor phs: {self.has_donor_phs} " \
                  f"({self.cbm_electron_occupation:5.3f} vs. {defaults.state_occupied_threshold})"
        return "\n".join([table, vbm_phs, cbm_phs])

    @staticmethod
    def _show_edge_info(edge_info: EdgeInfo, orb_diff: float):
        return [edge_info.band_idx + 1,
                f"{edge_info.energy:7.3f}",
                f"{edge_info.p_ratio:5.2f}",
                f"{edge_info.occupation:5.2f}",
                f"{orb_diff:5.2f}",
                pretty_orbital(edge_info.orbital_info.orbitals),
                pretty_coords(edge_info.kpt_coord)]


@dataclass
class BandEdgeStates(MSONable, ToJsonFileMixIn):
    states: List[BandEdgeState]  # by spin.

    @property
    def is_shallow(self):
        return any([i.is_shallow for i in self.states])

    @property
    def has_donor_phs(self):
        return any([i.has_donor_phs for i in self.states])

    @property
    def has_acceptor_phs(self):
        return any([i.has_acceptor_phs for i in self.states])

    @property
    def has_unoccupied_localized_state(self):
        return any([i.has_unoccupied_localized_state for i in self.states])

    @property
    def has_occupied_localized_state(self):
        return any([i.has_occupied_localized_state for i in self.states])

    @property
    def band_indices_from_vbm_to_cbm(self) -> List[int]:
        indices_set = set()
        for state in self.states:
            indices_set.add(state.vbm_info.band_idx)
            for lo in state.localized_orbitals:
                indices_set.add(lo.band_idx)
            indices_set.add(state.cbm_info.band_idx)
        return sorted([i for i in indices_set])

    def __str__(self):
        lines = [" -- band-edge states info"]
        for spin, state in zip(["up", "down"], self.states):
            lines.append(f"Spin-{spin}")
            lines.append(state.__str__())
            lines.append("")

        return "\n".join(lines)



# -*- coding: utf-8 -*-
#  Copyright (c) 2020. Distributed under the terms of the MIT License.

import pytest
from pydefect.defaults import defaults

from pydefect.input_maker.supercell_info import Interstitial, SupercellInfo
from pydefect.util.structure_tools import Coordination
from vise.tests.helpers.assertion import assert_msonable, assert_json_roundtrip


@pytest.fixture
def interstitial():
    return Interstitial(frac_coords=[0.25, 0.25, 0.25],
                        site_symmetry="m3m",
                        info="test")


def test_interstitial(interstitial):
    assert_msonable(interstitial)


def test_supercell_info_msonable(supercell_info):
    assert_msonable(supercell_info)


def test_supercell_info_from_dict(supercell_info):
    d = supercell_info.as_dict()
    d["transform_matrix"] = d.pop("transformation_matrix")
    expected = SupercellInfo.from_dict(d)
    assert expected == supercell_info


def test_supercell_info_distances(mocker, supercell_info):
    mock = mocker.patch("pydefect.util.structure_tools.defaults")
    mock.dist_tol = defaults.dist_tol
    mock.cutoff_distance_factor = 1.7
    assert supercell_info.coords("H1") == \
           Coordination({"H": [3.91], "He": [2.5, 3.0, 3.5]},
                        4.25, [1, 4, 5, 6])
    assert supercell_info.coords("He1") == \
           Coordination({"H": [2.5, 3.0, 3.5], "He": [3.91]},
                        4.25, [0, 2, 3, 7])


def test_supercell_info_interstitials_distances(mocker, supercell_info):
    mock = mocker.patch("pydefect.util.structure_tools.defaults")
    mock.dist_tol = defaults.dist_tol
    mock.cutoff_distance_factor = 1.7
    actual = supercell_info.interstitial_coords(0)
    expected = Coordination(distance_dict={'H': [2.62, 2.62, 2.62, 2.62],
                                           'He': [2.62, 2.62, 2.62, 2.62]},
                            cutoff=4.457,
                            neighboring_atom_indices=[0, 1, 2, 3, 4, 5, 6, 7])
    assert actual == expected


def test_supercell_info_str(mocker, supercell_info):
    mock = mocker.patch("pydefect.util.structure_tools.defaults")
    mock.dist_tol = defaults.dist_tol
    mock.cutoff_distance_factor = 1.7
    expected = """Space group: Fmmm
Transformation matrix: [1, 0, 0]  [0, 1, 0]  [0, 0, 1]
Cell multiplicity: 1

   Irreducible element: H1
        Wyckoff letter: a
         Site symmetry: mmm
         Cutoff radius: 4.25
          Coordination: {'H': [3.91], 'He': [2.5, 3.0, 3.5]}
      Equivalent atoms: 0..3
Fractional coordinates: 0.0000000  0.0000000  0.0000000
     Electronegativity: 2.2
       Oxidation state: 1

   Irreducible element: He1
        Wyckoff letter: b
         Site symmetry: mmm
         Cutoff radius: 4.25
          Coordination: {'H': [2.5, 3.0, 3.5], 'He': [3.91]}
      Equivalent atoms: 4..7
Fractional coordinates: 0.0000000  0.0000000  0.5000000
     Electronegativity: None
       Oxidation state: 0

-- interstitials
#1
                  Info: test
Fractional coordinates: 0.2500000  0.2500000  0.2500000
         Site symmetry: yy
          Coordination: {'H': [2.62, 2.62, 2.62, 2.62], 'He': [2.62, 2.62, 2.62, 2.62]}
"""
    assert str(supercell_info) == expected


def test_json_round_trip(supercell_info, tmpdir):
    assert_json_roundtrip(supercell_info, tmpdir)





# -*- coding: utf-8 -*-
#  Copyright (c) 2020 Kumagai group.
from pydefect.analyzer.defect_structure_info import DefectStructureInfo
from vise.analyzer.vesta.vesta_file import VestaFile
from pydefect.defaults import defaults
from pymatgen.core import Structure, DummySpecies, Element
from vise.util.typing import GenCoords


class MakeDefectVestaFile:
    def __init__(self,
                 defect_structure_info: DefectStructureInfo,
                 cutoff: float = defaults.show_structure_cutoff,
                 min_displace_w_arrows: float = 0.1,
                 arrow_factor: float = 3.0,
                 title: str = None):
        self.defect_structure = defect_structure_info.shifted_final_structure
        self.lattice = self.defect_structure.lattice

        self.removed_by_sub = defect_structure_info.site_diff.removed_by_sub
        self.inserted_by_sub = defect_structure_info.site_diff.inserted_by_sub
        self.removed = defect_structure_info.site_diff.removed
        self.inserted = defect_structure_info.site_diff.inserted
        self.removed_from_init = defect_structure_info.site_diff_from_initial.removed
        self.is_complex = defect_structure_info.site_diff.is_complex_defect

        self.center = defect_structure_info.center
        self.displacements = defect_structure_info.displacements

        self.cutoff = cutoff
        self.min_displace_w_arrows = min_displace_w_arrows
        self.arrow_factor = arrow_factor

        self._make_local_species_info()
        self._initial_local = self.make_initial_structure()
        self._final_local = self.make_final_structure()

        boundary = sum([[x - 0.5, x + 0.5]
                        for x in defect_structure_info.center], [])

        self.initial_vesta = \
            VestaFile(self._initial_local, title, boundary=boundary, show_label=False)
        self.final_vesta = \
            VestaFile(self._final_local, title, vectors=self._vectors,
                      vector_colors=self._vector_colors, boundary=boundary)

    def make_initial_structure(self):
        result = Structure(self.lattice, self._species, self._initial_coords)
        for _, specie, coord in self.removed_by_sub + self.removed:
            result.append(specie, coords=coord, properties={"name": f"{specie}"})
        fold_coords_in_structure(result, self.center)
        return result

    def make_final_structure(self):
        result = Structure(self.lattice, self._species, self._final_coords,
                           site_properties={"name": self._site_info})
        # inserted
        for _, specie, coords in self.inserted:
            result.append(specie, coords, properties={"name": f"{specie}_i"})
        # substituted
        for (_, i_specie, i_fcoord), (_, r_specie, r_fcoord) \
                in zip(self.inserted_by_sub, self.removed_by_sub):
            result.append(i_specie, i_fcoord,
                          properties={"name": f"{i_specie}_{r_specie}"})
        # removed. DummySpecies come later.
        for _, specie, coords in self.removed:
            result.append(DummySpecies(), coords,
                          properties={"name": f"Va_{specie}"})

        for _, specie, coords in self.removed_from_init:
            result.append(DummySpecies(), coords,
                          properties={"name": f"Initial_{specie}"})

        if self.is_complex:
            result.append(DummySpecies(), self.center,
                          properties={"name": "center"})

        fold_coords_in_structure(result, self.center)
        return result

    def _make_local_species_info(self):
        self._species, self._initial_coords, self._final_coords = [], [], []
        self._site_info, self._vectors, self._vector_colors = [], {}, []

        specie_idx = 1
        for specie, disp in zip(self.defect_structure, self.displacements):
            if disp and disp.distance_from_defect < self.cutoff:
                self._species.append(Element(disp.specie))
                self._initial_coords.append(disp.original_pos)
                self._final_coords.append(specie.frac_coords)
                info = f"{round(disp.distance_from_defect, 1)}"

                if disp and disp.displace_distance > self.min_displace_w_arrows:
                    self._vectors[specie_idx] = \
                        (v * self.arrow_factor for v in disp.disp_vector)
                    # max=200 instead of 255 since (255, 255, 255) is white.
                    rgb = int(200 * disp.angle / 180)
                    self._vector_colors.append((rgb, rgb, 100))
                    info += f"_{round(disp.displace_distance, 2)}"

                self._site_info.append(info)
                specie_idx += 1


def fold_coords_in_structure(structure: Structure, center: GenCoords) -> None:
    for site in structure:
        _, image = site.distance_and_image_from_frac_coords(center)
        site.frac_coords -= image


# -*- coding: utf-8 -*-
#  Copyright (c) 2020. Distributed under the terms of the MIT License.
from pathlib import Path

import pytest
from monty.serialization import loadfn
from pydefect.corrections.efnv_correction import \
    ExtendedFnvCorrection, PotentialSite
from pydefect.corrections.site_potential_plotter import \
    SitePotentialMplPlotter, SitePotentialPlotlyPlotter
from vise.util.dash_helper import show_png

try:
    import psutil
    PSUTIL_NOT_PRESENT = False
except ModuleNotFoundError:
    PSUTIL_NOT_PRESENT = True


@pytest.fixture
def efnv_cor():
    return ExtendedFnvCorrection(
        charge=1,
        point_charge_correction=0.0,
        defect_region_radius=2.8,
        sites=[
            PotentialSite(specie="H", distance=1.0, potential=-4, pc_potential=None),
            PotentialSite(specie="H", distance=2.0, potential=-3, pc_potential=None),
            PotentialSite(specie="He", distance=3.0, potential=-2,
                          pc_potential=-3),
            PotentialSite(specie="He", distance=4.0, potential=-1, pc_potential=-2)
        ],
        defect_coords=(0.0, 0.0, 0.0))


@pytest.mark.skipif(PSUTIL_NOT_PRESENT, reason="skipped for circle CI")
def test_site_potential_plotter_with_simple_example(efnv_cor):
    plotter = SitePotentialMplPlotter.from_efnv_corr(
        title="ZnO Va_O1_2", efnv_correction=efnv_cor)
    plotter.construct_plot()
    plotter.plt.show()


@pytest.mark.skipif(PSUTIL_NOT_PRESENT, reason="skipped for circle CI")
def test_site_potential_plotter_with_actual_file():
    efnv_cor = loadfn(Path(__file__).parent / "NaCl_Va_Na_-1_correction.json")
    plotter = SitePotentialMplPlotter.from_efnv_corr(
        title="NaCl Va_Na_-1", efnv_correction=efnv_cor)
    plotter.construct_plot()
    plotter.plt.show()


@pytest.mark.skipif(PSUTIL_NOT_PRESENT, reason="psutil does not exist")
def test_defect_energies_plotly_actual_plot(efnv_cor):
    fig = SitePotentialPlotlyPlotter.from_efnv_corr(
        title="ZnO Va_O1_2", efnv_correction=efnv_cor).create_figure()
    show_png(fig)


@pytest.mark.skipif(PSUTIL_NOT_PRESENT, reason="skipped for circle CI")
def test_gkfo_site_potential_plotter():
    gkfo = loadfn(Path(__file__).parent / "gkfo_correction.json")
    plotter = SitePotentialMplPlotter.from_gkfo_corr(
        title="GKFO", gkfo_correction=gkfo)
    plotter.construct_plot()
    plotter.plt.show()





# -*- coding: utf-8 -*-
#  Copyright (c) 2020 Kumagai group.
import re
from typing import Dict, Any

from pymatgen.core import Element


def remove_digits(name: str) -> str:
    """ "O1" -> "O" """
    return ''.join([i for i in name if not i.isdigit()])


def only_digits(name: str) -> str:
    """ "O1" -> "1" """
    return ''.join([i for i in name if i.isdigit()])


elements = [str(e) for e in Element]
e_va = elements + ["Va"]
e_i = elements + ["i"]


def defect_mpl_name(name: str) -> str:
    """ "Va_O1" -> "$V_{{\\rm O}1}$"
        "Mg_i1" -> "${\\rm Mg}_{i1}$" """
    in_name, out_name = name.split("_")
    if in_name in elements:
        in_name = "{\\rm " + in_name + "}"
    elif in_name == "Va":
        in_name = "V"

    r_out_name = remove_digits(out_name)
    if r_out_name in elements:
        out_name = "{{\\rm " + r_out_name + "}" + only_digits(out_name) + "}"
    else:
        out_name = "{" + out_name + "}"

    return f"${in_name}_{out_name}$"


def defect_plotly_name(name: str) -> str:
    """ "Va_O1" -> "<i>V</i><sub>O1</sub>" """
    in_name, out_name = name.split("_")
    if in_name == "Va":
        in_name = "<i>V</i>"
    out_name = f"<sub>{out_name}</sub>"
    return f"{in_name}{out_name}"


def defect_plotly_full_name(fullname: str) -> str:
    """ "Va_O1_1" -> "<i>V</i><sub>O1</sub><sup>1</sup>" """
    name, charge = fullname.rsplit('_', 1)
    return f"{defect_plotly_name(name)}<sup>{charge}</sup>"


def typical_defect_name(name: str) -> bool:
    x = name.split("_")
    if len(x) == 2:
        _in, _out = x
        if _in in e_va and remove_digits(_out) in e_i:
            return True
    return False


def prettify_names(d: Dict[str, Any], style) -> Dict[str, Any]:
    result = {}
    out_names = [name.split("_")[1] for name in d.keys()]
    for name, v in d.items():
        in_name, out_name = name.split("_")
        r_out_name = remove_digits(out_name)
        out_name = r_out_name if f"{r_out_name}2" not in out_names else out_name
        _name = "_".join([in_name, out_name])
        if _name in result:
            raise ValueError("The prettified names are conflicted. "
                             "Change the defect names, please.")
        if style is None:
            pass
        elif style == "mpl":
            _name = defect_mpl_name(_name)
        elif style == "plotly":
            _name = defect_plotly_name(_name)
        else:
            raise ValueError(f"Style {style} is not adequate. Choose from mpl"
                             f"or plotly.")
        result[_name] = v
    return result


# def defect_html_title_name(fullname):
#     x = fullname.split("_")
#     if len(x) == 2:
#         in_name, out_name = x
#     elif len(x) == 3:
#         in_name, out_name, charge = x
#     else:
#         raise ValueError
#
#     if in_name == "Va":
#         in_name = html.I("V")
#     else:
#         in_name = html.Span(in_name)
#
#     result = [in_name, html.Sub(out_name)]
#     if len(x) == 3:
#         result.append(html.Sup(charge))
#     return result





# -*- coding: utf-8 -*-
from abc import abstractmethod, ABC
from itertools import cycle
from typing import Optional, List

import numpy as np
import plotly.express as px
import plotly.graph_objects as go
from matplotlib import pyplot as plt
from mpl_toolkits.mplot3d.art3d import Poly3DCollection
from pydefect.chem_pot_diag.chem_pot_diag import ChemPotDiag
from pymatgen.core import Composition
from vise.util.plotly_util import sort_coords, make_triangles
from vise.util.string import latexify


class CpdMplSettings:
    def __init__(self):
        # Avoid type 3 font, http://phyletica.org/matplotlib-fonts/
        # https://stackoverflow.com/questions/34387893/output-matplotlib-figure-to-svg-with-text-as-text-not-curves
        self.rc_params = {'pdf.fonttype':  42, 'ps.fonttype': 42,
                          'svg.fonttype': "none"}

        self.grid_2d = {"color": 'black',
                        "alpha": 0.8,  # grid opacity
                        "linestyle": 'dashed',
                        "linewidth": 0.5}

        # Default zorder
        # Patch / PatchCollection	1
        # Line2D / LineCollection	2
        # Text	3
        # Big number of zorder is needed as it increases with num of simplexes
        self.label = {"size": 'smaller', "zorder": 100}

        self.alphabet_label = {"color": "blue", "weight": "bold", "zorder": 101}

        self.target_comp_text_color = "blue"
        self.comp_text_color = "black"

        self.simplex_2d = {"color": "black", "linewidth": 2, "zorder": 1}


class ChemPotDiagMplPlotter(ABC):
    def __init__(self,
                 cpd: ChemPotDiag,
                 mpl_defaults: Optional[CpdMplSettings] = CpdMplSettings()):
        self.cpd = cpd

        self._mpl_defaults = mpl_defaults
        self._add_ax()

    @abstractmethod
    def _add_ax(self):
        pass

    def draw_diagram(self, title: Optional[str] = None) -> plt:
        plt.rcParams.update(self._mpl_defaults.rc_params)
        self._plot()
        self._add_alphabet_labels()
        self._set_title(title)
        self._set_lims()
        self._set_labels()
        self._set_grid()
        plt.tight_layout()

        return plt

    @abstractmethod
    def _set_lims(self):
        pass

    @abstractmethod
    def _set_labels(self):
        pass

    @abstractmethod
    def _set_grid(self):
        pass

    def _plot(self):
        for comp, position in self.cpd.comp_centers.items():
            self.draw_simplex(comp)
            self._ax.text(*position,
                          latexify(comp),
                          color=self._text_color(comp),
                          **self._mpl_defaults.label,
                          **self._text_kwargs)

    @abstractmethod
    def draw_simplex(self, formula):
        pass

    @property
    @abstractmethod
    def _text_kwargs(self):
        pass

    def _text_color(self, formula):
        if self.cpd.target_coords and formula == self.cpd.target:
            return self._mpl_defaults.target_comp_text_color
        return self._mpl_defaults.comp_text_color

    def _set_title(self, title):
        if not title:
            title = f"Chemical potential diagram of {self.cpd.chemical_system}"
        self._ax.set_title(title)

    def _add_alphabet_labels(self):
        if self.cpd.target:
            for label, vertex_coords in self.cpd.target_coords.items():
                self._ax.text(*transpose(vertex_coords), label,
                              **self._mpl_defaults.alphabet_label)


class ChemPotDiag2DMplPlotter(ChemPotDiagMplPlotter):

    def _add_ax(self):
        self._ax = plt.figure().add_subplot(111)

    @property
    def _text_kwargs(self):
        return {}

    def _set_lims(self):
        self._ax.set_xlim(self.cpd.min_range, 0)
        self._ax.set_ylim(self.cpd.min_range, 0)
        self._ax.set_aspect('equal', adjustable='box')

    def _set_labels(self):
        vertex_elements = self.cpd.vertex_elements
        self._ax.set_xlabel(f"Chemical potential of {vertex_elements[0]}")
        self._ax.set_ylabel(f"Chemical potential of {vertex_elements[1]}")

    def _set_grid(self):
        self._ax.grid(**self._mpl_defaults.grid_2d)

    def draw_simplex(self, formula):
        vertex_coords = self.cpd.polygons[formula]
        plt.plot(*transpose(vertex_coords), **self._mpl_defaults.simplex_2d)


class ChemPotDiag3DMplPlotter(ChemPotDiagMplPlotter):
    def _add_ax(self):
        self._ax = plt.figure().add_subplot(111, projection='3d')

    @property
    def _text_kwargs(self):
        return {"ha": "center", "va": "center"}

    def _set_lims(self):
        self._ax.set_xlim3d(self.cpd.min_range, 0)
        self._ax.set_ylim3d(0, self.cpd.min_range)
        self._ax.set_zlim3d(self.cpd.min_range, 0)

    def _set_labels(self):
        label = [f"Chemical potential of {e}" for e in self.cpd.vertex_elements]
        self._ax.set_xlabel(label[0])
        self._ax.set_ylabel(label[1])
        self._ax.set_zlabel(label[2])

    def _set_grid(self):
        pass

    def draw_simplex(self, formula):
        vertex_coords = self.cpd.polygons[formula]
        atomic_fractions = self.cpd.atomic_fractions(formula)

        face = Poly3DCollection([sort_coords(np.array(vertex_coords))])
        face.set_color(self._3d_simplex_color(atomic_fractions))
        face.set_edgecolor("black")
        self._ax.add_collection3d(face)

    @staticmethod
    def _3d_simplex_color(atomic_fractions):
        return [(f + 1.5) / 2.5 for f in atomic_fractions]


def transpose(target_list: List[List[float]]) -> List[List[float]]:
    return np.array(target_list).transpose().tolist()


class ChemPotDiagPlotlyPlotter(ABC):
    def __init__(self, cpd: ChemPotDiag):
        self.cpd = cpd
        self.axes_titles = [f"μ <sub>{el}</sub> (eV)"
                            for el in self.cpd.vertex_elements]
        title = f"Chemical potential diagram of {self.cpd.chemical_system}"
        self.range = [self.cpd.min_range * 1.001, -self.cpd.min_range * 0.1]
        self.common_layout = dict(title=title, width=700, height=700,
                                  title_font_size=30, showlegend=False)
        self.title_axis = dict(xaxis_title=self.axes_titles[0],
                               yaxis_title=self.axes_titles[1],
                               xaxis_range=self.range,
                               yaxis_range=self.range)


class ChemPotDiag2DPlotlyPlotter(ChemPotDiagPlotlyPlotter):
    def __init__(self, cpd: ChemPotDiag):
        super().__init__(cpd)
        self.hover = "<b>label %{text}</b><br> energy (%{x:.2f}, %{y:.2f})"

    @property
    def figure(self):
        fig = go.Figure()

        for comp, vertices in self.cpd.polygons.items():
            (x0, y0), (x1, y1) = vertices[0], vertices[1]
            x_ave, y_ave = (x0 + x1) / 2, (y0 + y1) / 2
            if x_ave == 0.0:
                pos = "middle right"
            else:
                pos = "top center"

            fig.add_shape(dict(type="line", x0=x0, y0=y0, x1=x1, y1=y1,
                               line=dict(color="RoyalBlue", width=3)))
            fig.add_trace(go.Scatter(x=[x_ave], y=[y_ave],
                                     text=[clean_formula(comp)],
                                     mode="text",
                                     textposition=pos,
                                     textfont=dict(color="black", size=24),
                                     hoverinfo='skip'))

        if self.cpd.target:
            x, y, text = [], [], []
            for label, cp in self.cpd.target_coords.items():
                x.append(cp[0])
                y.append(cp[1])
                text.append(label)
            fig.add_trace(go.Scatter(x=x, y=y, text=text, mode='markers+text',
                                     textposition="bottom left",
                                     textfont=dict(size=24),
                                     hovertemplate=self.hover))

        fig.update_traces(marker_size=15)
        fig.update_layout(font_size=24, **self.common_layout, **self.title_axis)
        return fig


class ChemPotDiag3DPlotlyPlotter(ChemPotDiagPlotlyPlotter):
    def __init__(self, cpd: ChemPotDiag):
        super().__init__(cpd)
        self.hover = \
            "<b>label %{text}</b><br> energy (%{x:.2f}, %{y:.2f}, %{z:.2f})"

    @property
    def figure(self):
        data = []
        color_cycle = cycle(px.colors.qualitative.Dark24)

        for comp, vertex_coords in self.cpd.polygons.items():
            color = next(color_cycle)
            vertex_coords = sort_coords(np.array(vertex_coords))
            # add surface
            d = make_triangles(vertex_coords)
            data.append(go.Mesh3d(**d, alphahull=-1, opacity=0.3,
                                  hoverinfo='skip', color=color))
            # add formula
            data.append(go.Scatter3d(x=[sum(d["x"]) / len(vertex_coords)],
                                     y=[sum(d["y"]) / len(vertex_coords)],
                                     z=[sum(d["z"]) / len(vertex_coords)],
                                     text=[clean_formula(comp)],
                                     mode="text",
                                     textposition="middle center",
                                     textfont=dict(color=color, size=24),
                                     hoverinfo='skip'))
        if self.cpd.target:
            target_coords = self.cpd.target_coords
            x = [cp[0] for cp in target_coords.values()]
            y = [cp[1] for cp in target_coords.values()]
            z = [cp[2] for cp in target_coords.values()]
            label = [key for key in target_coords]
            data.append(go.Scatter3d(x=x, y=y, z=z, text=label,
                                     mode='markers+text',
                                     textfont=dict(size=24),
                                     hovertemplate=self.hover))
        fig = go.Figure(data=data)
        # fig.update_traces(marker_size=15)
        fig.update_layout(scene=dict(zaxis_title=self.axes_titles[2],
                                     zaxis_range=self.range,
                                     **self.title_axis),
                          font_size=15,
                          **self.common_layout)

        return fig


def clean_formula(formula: str) -> str:
    s = []
    for char in formula:
        if char == "1" or char == " ":
            continue
        elif char.isdigit():
            s.append(f"<sub>{char}</sub>")
        else:
            s.append(char)

    return "".join(s)



# -*- coding: utf-8 -*-
#  Copyright (c) 2020. Distributed under the terms of the MIT License.
from itertools import groupby, cycle
from typing import Optional, List

import plotly.graph_objects as go
from matplotlib import pyplot as plt
from pydefect.corrections.efnv_correction import \
    ExtendedFnvCorrection, PotentialSite
from pydefect.corrections.gkfo_correction import GkfoCorrection
from pydefect.defaults import defaults
from vise.util.matplotlib import float_to_int_formatter


class PotentialPlotterMplSettings:
    def __init__(self,
                 colors: Optional[List[str]] = None,
                 line_width: float = 1.0,
                 radius_line_width: float = 1.0,
                 radius_line_color: str = "black",
                 radius_line_style: str = "-.",
                 ave_pot_diff_line_width: float = 1.5,
                 ave_pot_diff_line_color: str = "red",
                 ave_pot_diff_line_style: str = ":",
                 circle_size: int = 15,
                 zero_line_width: float = 0.5,
                 zero_line_color: str = "black",
                 zero_line_style: str = ":",
                 tick_label_size: Optional[int] = 12,
                 title_font_size: Optional[int] = 15,
                 label_font_size: Optional[int] = 15):
        self.colors = cycle(colors) if colors else defaults.defect_energy_colors
        self.line_width = line_width
        self.circle_size = circle_size
        self.tick_label_size = tick_label_size
        self.title_font_size = title_font_size
        self.label_font_size = label_font_size

        self.vline = {"linewidth": radius_line_width,
                      "color": radius_line_color,
                      "linestyle": radius_line_style}

        self.hline = {"linewidth": ave_pot_diff_line_width,
                      "color": ave_pot_diff_line_color,
                      "linestyle": ave_pot_diff_line_style}

        self.zero_line = {"linewidth": zero_line_width,
                          "color": zero_line_color,
                          "linestyle": zero_line_style}


class SitePotentialPlotter:
    def __init__(self,
                 title: str,
                 sites: List[PotentialSite],
                 defect_region_radius: float,
                 ave_pot_diff: float,
                 x_unit: Optional[str] = "Å",
                 y_unit: Optional[str] = "V"):
        self._title = title
        self.sites = sites
        self.defect_region_radius = defect_region_radius
        self.ave_pot_diff = ave_pot_diff

        self._x_unit = x_unit
        self._y_unit = y_unit

    @classmethod
    def from_efnv_corr(cls,
                       title,
                       efnv_correction: ExtendedFnvCorrection,
                       x_unit: Optional[str] = "Å",
                       y_unit: Optional[str] = "V"):
        return cls(title=title,
                   sites=efnv_correction.sites,
                   defect_region_radius=efnv_correction.defect_region_radius,
                   ave_pot_diff=efnv_correction.average_potential_diff,
                   x_unit=x_unit, y_unit=y_unit)

    @classmethod
    def from_gkfo_corr(cls,
                       title,
                       gkfo_correction: GkfoCorrection,
                       x_unit: Optional[str] = "Å",
                       y_unit: Optional[str] = "V"):
        efnv = gkfo_correction.init_efnv_correction
        ave_pot_diff = gkfo_correction.average_potential_diff_by_addition
        return cls(title=title,
                   sites=gkfo_correction.gkfo_sites,
                   defect_region_radius=efnv.defect_region_radius,
                   ave_pot_diff=ave_pot_diff,
                   x_unit=x_unit, y_unit=y_unit)


class SitePotentialPlotlyPlotter(SitePotentialPlotter):
    def create_figure(self):
        common = {"mode": "markers", "marker_size": 10}
        fig = go.Figure()
        fig.update_layout(
            title="Potential",
            xaxis_title=f"Distance ({self._x_unit})",
            yaxis_title=f"Potential ({self._y_unit})",
            title_font_size=30,
            font_size=24, width=800, height=700)
        all_distances = []
        pc_pot_distances = []
        all_potentials = []
        pc_potentials = []
        potential_diffs = []
        key = lambda x: x.specie
        for _, grouped_sites in groupby(sorted(self.sites, key=key), key=key):
            distances = []
            potentials = []
            atom_idxs = []
            for site in grouped_sites:
                distances.append(site.distance)
                all_distances.append(site.distance)
                all_potentials.append(site.potential)
                potentials.append(site.potential)
#                atom_idxs.append(site.)

                if site.pc_potential:
                    pc_pot_distances.append(site.distance)
                    pc_potentials.append(site.pc_potential)
                    potential_diffs.append(site.diff_pot)

            fig.add_trace(go.Scatter(x=distances, y=potentials,
                                     text=atom_idxs,
                                     name=site.specie,
                                     marker_symbol="circle",
                                     **common))

        min_y = min(all_potentials + pc_potentials + potential_diffs) - 0.4
        max_y = max(all_potentials + pc_potentials + potential_diffs) + 0.4
        max_x = max(all_distances) + 0.4
        fig.add_shape(dict(type="line",
                           x0=self.defect_region_radius,
                           x1=self.defect_region_radius,
                           y0=min_y, y1=max_y,
                           line=dict(color="royalblue", width=1.5, dash="dot")))
        fig["layout"]["xaxis"]["range"] = [0, max_x]
        fig["layout"]["yaxis"]["range"] = [min_y, max_y]

        fig.add_trace(go.Scatter(x=pc_pot_distances, y=pc_potentials,
                                 name="PC", marker_symbol="diamond",
                                 **common))
        fig.add_trace(go.Scatter(x=pc_pot_distances, y=potential_diffs,
                                 name="Diff", marker_symbol="cross",
                                 **common))

        fig.add_shape(dict(type="line",
                           x0=self.defect_region_radius, x1=max_x,
                           y0=self.ave_pot_diff, y1=self.ave_pot_diff,
                           line=dict(color="red", width=1.5, dash="dot"),
                           name="alignment"))

        return fig


class SitePotentialMplPlotter(SitePotentialPlotter):
    def __init__(self, **kwargs):
        super().__init__(**kwargs)
        self._mpl_defaults = kwargs.get("mpl_defaults",
                                        PotentialPlotterMplSettings())
        self.plt = plt

    def construct_plot(self):
        self._add_potentials()
        self._add_lines()
        self._set_labels()
        self._set_x_range()
        self._set_title()
        self._set_formatter()
        self.plt.tight_layout()
#        self.plt.gca().legend(bbox_to_anchor=(1, 0.5), loc='center left')

    def _add_potentials(self):
        all_distances = []
        pc_pot_distances = []
        pc_potentials = []
        potential_diffs = []
        key = lambda x: x.specie
        for _, grouped_sites in groupby(sorted(self.sites, key=key), key=key):
            distances = []
            potentials = []
            for site in grouped_sites:
                distances.append(site.distance)
                all_distances.append(site.distance)
                potentials.append(site.potential)

                if site.pc_potential:
                    pc_pot_distances.append(site.distance)
                    pc_potentials.append(site.pc_potential)
                    potential_diffs.append(site.diff_pot)

            color = next(self._mpl_defaults.colors)
            self.plt.scatter(distances, potentials, marker="o", color=color,
                             label=str(site.specie))

        self._max_distance = max(all_distances)

        self.plt.scatter(pc_pot_distances, pc_potentials, marker="1",
                         color="b", label="point charge")
        self.plt.scatter(pc_pot_distances, potential_diffs, marker="+",
                         color="r", label="potential difference")

        self.plt.legend(bbox_to_anchor=(0, 1.05), loc='upper left',
                        borderaxespad=0, fontsize=8)

    def _add_lines(self):
        self.plt.axvline(x=self.defect_region_radius,
                         **self._mpl_defaults.vline)
        self.plt.plot([self.defect_region_radius, self._max_distance * 1.1],
                      [self.ave_pot_diff, self.ave_pot_diff],
                      **self._mpl_defaults.hline)
        self.plt.axhline(y=0, **self._mpl_defaults.zero_line)

    def _set_x_range(self):
        self.plt.xlim(0, self._max_distance * 1.05)

    def _set_title(self):
        self.plt.title(self._title, size=self._mpl_defaults.title_font_size)

    def _set_labels(self):
        self.plt.xlabel(f"Distance from a defect ({self._x_unit})",
                        size=self._mpl_defaults.label_font_size)
        self.plt.ylabel(f"Potential ({self._y_unit})",
                        size=self._mpl_defaults.label_font_size)

    def _set_formatter(self):
        axis = self.plt.gca()
        axis.xaxis.set_major_formatter(float_to_int_formatter)
        axis.yaxis.set_major_formatter(float_to_int_formatter)
        axis.tick_params(labelsize=self._mpl_defaults.tick_label_size)


# -*- coding: utf-8 -*-
#  Copyright (c) 2020 Kumagai group.
from pydefect.analyzer.calc_results import CalcResults
from pydefect.analyzer.defect_energy import DefectEnergy, DefectEnergyInfo
from pydefect.analyzer.make_defect_energy_info import make_defect_energy_info, \
    num_atom_differences
from pydefect.chem_pot_diag.chem_pot_diag import StandardEnergies
from pydefect.corrections.abstract_correction import Correction
from pydefect.input_maker.defect_entry import DefectEntry
from pymatgen.core import IStructure, Lattice


def test_make_defect_energy_info(mocker):
    defect_entry = mocker.Mock(DefectEntry, autospec=True)
    defect_entry.name = "Va_Mg1"
    defect_entry.charge = -1

    calc_results = mocker.Mock(CalcResults, autospec=True)
    calc_results.structure = IStructure(Lattice.cubic(1.0), ["O"], [[0.0]*3])
    calc_results.energy = 10.0
    calc_results.electronic_conv = False

    correction = mocker.Mock(Correction, autospec=True)
    correction.correction_dict = {"a": 10.0}

    p_calc_results = mocker.Mock(CalcResults, autospec=True)
    p_calc_results.structure = IStructure(Lattice.cubic(1.0),
                                          ["Mg", "O"], [[0.0]*3]*2)
    p_calc_results.energy = 1.0

    standard_energies = StandardEnergies({"Mg": 10.0, "O": 20.0})

    unitcell = mocker.Mock()
    unitcell.vbm = 100.0

    actual = make_defect_energy_info(defect_entry, calc_results, correction,
                                     p_calc_results, standard_energies, unitcell)
    energy = DefectEnergy(formation_energy=10.0 - 1.0 + 10 - 100.0,
                          energy_corrections={"a": 10.0},
                          is_shallow=None)
    expected = DefectEnergyInfo(name="Va_Mg1", charge=-1,
                                atom_io={"Mg": -1}, defect_energy=energy)
    assert actual == expected


def test_num_atom_diff():
    s1 = IStructure(Lattice.cubic(1), ["H", "He"], [[0] * 3] * 2)
    s2 = IStructure(Lattice.cubic(1), ["H", "Li"], [[0] * 3] * 2)
    assert num_atom_differences(s1, s2) == {"He": 1, "Li": -1}


# -*- coding: utf-8 -*-
#  Copyright (c) 2020. Distributed under the terms of the MIT License.
import traceback
from pathlib import Path
from typing import List, Callable, Any

import numpy as np
from vise.util.logger import get_logger

logger = get_logger(__name__)


def sanitize_matrix(matrix: List[int]) -> List[List[int]]:

    if len(matrix) == 9:
        return [matrix[:3], matrix[3:6], matrix[6:]]
    elif len(matrix) == 3:
        result = np.eye(3, dtype=int)
        for i in range(3):
            result[i, i] = matrix[i]
        return result.tolist()
    elif len(matrix) == 1:
        result = np.eye(3, dtype=int)
        for i in range(3):
            result[i, i] = matrix[0]
        return result.tolist()
    else:
        raise ValueError(f"Matrix element length {len(matrix)} is improper.")


def str_int_to_int(x):
    try:
        return int(x)
    except ValueError:
        return x


def parse_dirs(dirs: List[Path],
               _inner_function: Callable[[Path], Any],
               verbose: bool = False,
               output_filename: str = None):
    failed_directories = []
    parsed_results = []
    for _dir in dirs:
        if _dir.is_file():
            logger.info(f"{_dir} is a file, so skipped.")
            continue
        if output_filename and (_dir / output_filename).exists():
            logger.info(f"In {_dir}, {output_filename} already exists.")
            continue

        logger.info(f"Parsing data in {_dir} ...")
        try:
            _return = _inner_function(_dir)
            if _return:
                parsed_results.append(_return)
        except Exception as e:
            if verbose:
                print(traceback.print_exc())
            else:
                print(e.args[1])
            logger.warning(f"Failing parsing {_dir} ...")
            failed_directories.append(str(_dir))
            continue
    if failed_directories:
        failed_dir_string = '\n'.join(failed_directories)
        logger.warning(f"Failed directories are:\n{failed_dir_string}")
        logger.warning(f"To see details, try to run with --verbose")
    else:
        logger.info("Parsing all the directories succeeded.")

    return parsed_results or None


import os

from pydefect import __version__
from setuptools import setup, find_packages

cmdclass = {}
ext_modules = []

module_dir = os.path.dirname(os.path.abspath(__file__))
reqs_raw = open(os.path.join(module_dir, "requirements.txt")).read()
reqs_list = [r.replace("==", "~=") for r in reqs_raw.split("\n")]

#with open("README.md", "r") as fh:
#    long_description = fh.read()

setup(
    name='pydefect',
    version=__version__,
    author='Yu Kumagai',
    author_email='yuuukuma@gmail.com',
    url='https://github.com/kumagai-group/pydefect',
    packages=find_packages(),
    license='MIT license',
    description="Integrated environment for first-principles point-defect "
                "calculations",
    classifiers=[
        'Programming Language :: Python :: 3.7',
        "License :: OSI Approved :: MIT License",
    ],
    install_requires=reqs_list,
    cmdclass=cmdclass,
    ext_modules=ext_modules,
    include_package_data=True,
    entry_points={
        'console_scripts': [
            'pydefect = pydefect.cli.main:main',
            'pydefect_util = pydefect.cli.main_util:main',
            'pydefect_vasp = pydefect.cli.vasp.main_vasp:main',
            'pydefect_vasp_util = pydefect.cli.vasp.main_vasp_util:main',
            'pydefect_print = pydefect.cli.main_print_json:main'
        ]
    }
)



# -*- coding: utf-8 -*-
#  Copyright (c) 2020. Distributed under the terms of the MIT License.
import logging
from pathlib import Path

import pytest

from pydefect.cli.main_tools import sanitize_matrix, str_int_to_int, parse_dirs


def test_sanitize_matrix_9_input_values():
    actual = sanitize_matrix(list(range(9)))
    expected = [[0, 1, 2], [3, 4, 5], [6, 7, 8]]
    assert actual == expected


def test_sanitize_matrix_3_input_values():
    actual = sanitize_matrix(list(range(1, 4)))
    expected = [[1, 0, 0], [0, 2, 0], [0, 0, 3]]
    assert actual == expected


def test_sanitize_matrix_1_input_value():
    actual = sanitize_matrix([4])
    expected = [[4, 0, 0], [0, 4, 0], [0, 0, 4]]
    assert actual == expected


def test_sanitize_matrix_raise_error_for_improper_matrix_element_length():
    with pytest.raises(ValueError):
         sanitize_matrix([1, 4])


def test_str_int_to_int():
    assert str_int_to_int("a") == "a"
    assert str_int_to_int("1") == 1
    assert str_int_to_int("1.1") == "1.1"


"""
TODO
- return matrix

DONE
"""


def print_a_to_file_x(path: Path):
    filename = path / "x"
    filename.write_text("a")


def test_parse_dirs(tmpdir):
    pathlib_tmpdir = Path(tmpdir)
    parse_dirs([Path(tmpdir)], _inner_function=print_a_to_file_x)

    actual = pathlib_tmpdir / "x"
    assert actual.read_text() == "a"


def test_parse_dirs_file(tmpdir):
    tmpfile = Path(tmpdir) / "file"
    tmpfile.touch()
    parse_dirs([tmpfile], _inner_function=print_a_to_file_x)


def test_parse_dirs_fail(tmpdir, capsys):
    parse_dirs([Path(tmpdir) / "y"], _inner_function=print_a_to_file_x)
    captured = capsys.readouterr()
    assert captured.out == "No such file or directory\n"


def test_parse_dirs_skip_when_output_file_exist(tmpdir):
    LOGGER = logging.getLogger(__name__)
    LOGGER.info('Testing now.')
    (Path(tmpdir) / "x").touch()
    parse_dirs([Path(tmpdir)], _inner_function=print_a_to_file_x,
               output_filename="x")



# -*- coding: utf-8 -*-
#  Copyright (c) 2020 Kumagai group.
import sys

from monty.serialization import loadfn


def main():
    lines = [loadfn(filename).__str__() for filename in sys.argv[1:]]
    print("\n\n".join(lines))


if __name__ == "__main__":
    main()



# -*- coding: utf-8 -*-
#  Copyright (c) 2020 Kumagai group.
import argparse
import sys

from monty.serialization import loadfn
from pydefect.chem_pot_diag.chem_pot_diag import ChemPotDiag
from pydefect.chem_pot_diag.cpd_plotter import ChemPotDiag3DPlotlyPlotter, \
    ChemPotDiag2DPlotlyPlotter


def parse_args(args):
    parser = argparse.ArgumentParser(description="")
    parser.add_argument("-c", "--chem_pot_diag", dest="cpd", type=loadfn)
    parser.add_argument("--port", type=int)
    return parser.parse_args(args)


if __name__ == "__main__":
    args = parse_args(sys.argv[1:])
    cpd: ChemPotDiag = args.cpd
    if cpd.dim == 2:
        plotter = ChemPotDiag2DPlotlyPlotter(args.cpd)
    elif cpd.dim == 3:
        plotter = ChemPotDiag3DPlotlyPlotter(args.cpd)
    else:
        raise ValueError("Dimension must be 2 or 3.")
    fig = plotter.figure
    fig.show()
#    app.run_server(debug=True, port=args.port)


# -*- coding: utf-8 -*-
#  Copyright (c) 2020. Distributed under the terms of the MIT License.
from collections import defaultdict
from dataclasses import dataclass
from typing import List, Dict, Optional, Set

import numpy as np
from monty.json import MSONable
from pydefect.defaults import defaults
from pymatgen.core import Structure, Element


class Distances:
    def __init__(self,
                 structure: Structure,
                 center_coord: np.array,
                 dist_tol: float = None):
        self.structure = structure
        self.coord = center_coord
        self.dist_tol = dist_tol or defaults.dist_tol

    def distances(self, remove_self=True, specie=None) -> List[float]:
        result = []
        lattice = self.structure.lattice
        for site in self.structure:
            if specie and Element(specie) != site.specie:
                result.append(float("inf"))
                continue
            distance, _ = \
                lattice.get_distance_and_image(site.frac_coords, self.coord)
            if remove_self and distance < 1e-5:
                continue
            result.append(distance)

        return result

    def atom_idx_at_center(self, specie: str) -> Optional[int]:
        distances = self.distances(remove_self=False, specie=specie)
        sorted_dists = sorted(distances)
        if sorted_dists[0] > self.dist_tol:
            return None
        return np.argmin(distances)

    @property
    def shortest_distance(self) -> float:
        return min(self.distances())

    def coordination(self, include_on_site=False, cutoff_factor=None
                     ) -> "Coordination":
        cutoff_factor = cutoff_factor or defaults.cutoff_distance_factor
        cutoff = self.shortest_distance * cutoff_factor
        elements = [element.specie.name for element in self.structure]
        e_d = zip(elements, self.distances(remove_self=False))

        unsorted_distances = defaultdict(list)
        neighboring_atom_indices = []
        for i, (element, distance) in enumerate(e_d):
            if distance < cutoff:
                if include_on_site or \
                        (include_on_site is False and distance > 1e-5):
                    unsorted_distances[element].append(round(distance, 2))
                    neighboring_atom_indices.append(i)

        distance_dict = {}
        for element, distances in unsorted_distances.items():
            distance_dict[element] = sorted(distances)

        return Coordination(distance_dict, round(cutoff, 3),
                            neighboring_atom_indices)


@dataclass
class Coordination(MSONable):
    distance_dict: Dict[str, List]
    cutoff: float
    neighboring_atom_indices: List[int]
